import loggingbootstrap
import logging
import traceback
import sys
import os
from pathlib import Path

# IMPORTANT: used to import the user_settings.py file
sys.path.insert(0, os.getcwd())

try:
    from confapp import conf

    # Initiating logging for pyforms. It has to be initiated manually here because we don't know yet
    # the logger filename as specified on settings
    loggingbootstrap.create_double_logger("pyforms", logging.INFO, 'app.log', logging.INFO)

except ImportError as err:
    logging.getLogger().critical(str(err), exc_info=True)
    exit("Could not load pyforms! Is it installed?")

try:
    # pyforms is imported here first time through pyforms
    try:
        import user_settings
    except ModuleNotFoundError:
        home = str(Path.home())
        home_settings_file = os.path.join(home, 'user_settings.py')

        if not os.path.isfile(home_settings_file):
            with open(home_settings_file, 'w') as out:
                out.write("SETTINGS_PRIORITY = 0\n\n")
                out.write("GENERIC_EDITOR_PLUGINS_LIST = ['pybpodgui_plugin']")

        sys.path.insert(0, home)
        import user_settings

    conf += user_settings

    from pyforms_generic_editor import settings

    loggingbootstrap.create_double_logger(
        "pyforms_generic_editor",
        conf.APP_LOG_HANDLER_CONSOLE_LEVEL,
        conf.APP_LOG_FILENAME,
        conf.APP_LOG_HANDLER_FILE_LEVEL
    )

    loggingbootstrap.create_double_logger(
        "pyforms",
        conf.APP_LOG_HANDLER_CONSOLE_LEVEL,
        conf.APP_LOG_FILENAME,
        conf.APP_LOG_HANDLER_FILE_LEVEL
    )

    # pyforms.controls is imported here first time
    from pyforms_generic_editor.editor.base_editor import BaseEditor as Editor

except Exception as err:
    exc_type, exc_value, exc_traceback = sys.exc_info()
    logging.getLogger("pyforms_generic_editor").critical(str(err), exc_info=True)
    conf.GENERIC_EDITOR_LOAD_EXCEPTION_TRACEBACK = traceback.format_exc()
    conf.GENERIC_EDITOR_LOAD_EXCEPTION_LINE = exc_traceback.tb_lineno
    settings.GENERIC_EDITOR_TITLE = 'PLEASE EDIT USER SETTINGS AND RESTART APP'

    from pyforms_generic_editor.editor.safe_mode_editor import SafeModeEditor as Editor


def start():
    import pyforms
    pyforms.start_app(Editor, conf.GENERIC_EDITOR_WINDOW_GEOMETRY)


if __name__ == '__main__':
    start()



# !/usr/bin/python3
# -*- coding: utf-8 -*-
import logging

from confapp import conf

from AnyQt.QtGui import QIcon

from pybpodgui_plugin.models.setup.board_task.board_task_window import BoardTaskWindow

logger = logging.getLogger(__name__)


class BoardTaskUIBusy(BoardTaskWindow):
    """
    TODO
    """

    def __init__(self, setup):
        super(BoardTaskUIBusy, self).__init__(setup)
        self.__running_icon = QIcon(conf.PLAY_SMALL_ICON)

    def update_ui(self):
        pass



# !/usr/bin/python3
# -*- coding: utf-8 -*-

import logging
import os
import shutil

import pandas as pd

from pybpodapi.com.messaging.session_info import SessionInfo
from pybpodapi.session import Session
from pybpodgui_api.com.messaging.parser import BpodMessageParser
from pybpodgui_api.models.session.session_base import SessionBase
from sca.formats import csv

from pybpodapi.utils import date_parser

logger = logging.getLogger(__name__)


class SessionIO(SessionBase):
    """

    """

    def __init__(self, setup):
        super(SessionIO, self).__init__(setup)

        # initial name. Used to track if the name was updated
        self.initial_name = None

    ##########################################################################
    ####### FUNCTIONS ########################################################
    ##########################################################################

    def collect_data(self, data):
        data.update({'name': self.name})
        data.update({'uuid4': self.uuid4})
        data.update({'started': str(self.started.strftime('%Y%m%d-%H%M%S')) if self.started else None})
        data.update({'ended': str(self.ended.strftime('%Y%m%d-%H%M%S')) if self.ended else None})
        data.update({'setup': str(self.setup.uuid4)})
        data.update({'task': str(self.task.uuid4 if self.task else None)})
        data.update({'board': str(self.setup.board.uuid4 if self.setup.board else None)})
        data.update({'serial_port': self.setup.board.serial_port})

        data.update({'subjects': []})
        for subject in self.subjects:
            data['subjects'].append(subject)

        data.update({'variables': []})
        for var in self.variables:
            data['variables'].append(var)

        return data

    def save(self):
        """

        :param parent_path:
        :return:
        """
        if not self.name:
            logger.warning("Skipping session without name")
        else:
            if self.initial_name is not None:
                initial_path = os.path.join(self.setup.path, 'sessions', self.initial_name)

                if initial_path != self.path:
                    shutil.move(initial_path, self.path)
                    current_filepath = os.path.join(self.path, self.initial_name+'.csv')
                    future_filepath = os.path.join(self.path, self.name+'.csv')
                    shutil.move(current_filepath, future_filepath)

            self.initial_name = self.name

    def load(self, path):
        """

        :param session_path:
        :param data:
        :return:
        """
        self.name = os.path.basename(path)
        # only set the filepath if it exists
        filepath = os.path.join(self.path, self.name+'.csv')

        if not os.path.exists(filepath):
            return

        try:
            self.filepath = filepath
            csvreader = self.load_info()
            self.uuid4 = csvreader.uuid4
        except FileNotFoundError:
            logger.warning('File not found: '+filepath)
            self.filepath = None

    def load_contents(self, init_func=None, update_func=None, end_func=None):
        """
        Parses session history file, line by line and populates the history message on memory.
        """
        if not self.filepath:
            return

        nrows = csv.reader.count_metadata_rows(self.filepath)

        with open(self.filepath) as filestream:
            self.data = pd.read_csv(filestream,
                                    delimiter=csv.CSV_DELIMITER,
                                    quotechar=csv.CSV_QUOTECHAR,
                                    quoting=csv.CSV_QUOTING,
                                    skiprows=nrows,
                                    memory_map=True
                                    )

        res = self.data.query("MSG=='{0}'".format(Session.INFO_SESSION_ENDED))
        for index, row in res.iterrows():
            self.ended = date_parser.parse(row['+INFO'])

        res = self.data.query("TYPE in ['VAL', 'TRIAL'] or MSG=='SESSION-ENDED'")
        variables = []
        for index, row in res.iterrows():
            if row['TYPE'] == 'TRIAL':
                variables.append(['New trial', None])
            elif row['MSG'] == 'SESSION-ENDED':
                variables.append(['Session ended', None])
            else:
                variables.append([row['MSG'], row['+INFO']])
        #self.variables = variables

    def load_info(self):
        if not self.filepath:
            return

        with open(self.filepath) as filestream:
            csvreader = csv.reader(filestream)
            self.subjects = []

            count = 0
            for row in csvreader:
                msg = BpodMessageParser.fromlist(row)

                if msg:
                    if isinstance(msg, SessionInfo):
                        if msg.infoname == Session.INFO_SESSION_NAME:
                            self.task_name = msg.infovalue

                        elif msg.infoname == Session.INFO_CREATOR_NAME:
                            self.creator = msg.infovalue

                        elif msg.infoname == Session.INFO_SESSION_STARTED:
                            self.started = date_parser.parse(msg.infovalue)

                        elif msg.infoname == Session.INFO_SESSION_ENDED:
                            self.ended = date_parser.parse(msg.infovalue)

                        elif msg.infoname == Session.INFO_SERIAL_PORT:
                            self.board_serial_port = msg.infovalue

                        elif msg.infoname == Session.INFO_BOARD_NAME:
                            self.board_name = msg.infovalue

                        elif msg.infoname == Session.INFO_SETUP_NAME:
                            self.setup_name = msg.infovalue

                        elif msg.infoname == Session.INFO_SUBJECT_NAME:
                            self.subjects += [msg.infovalue]
                            name, uuid4 = eval(msg.infovalue)
                            subj = self.project.find_subject_by_id(uuid4)
                            if subj is not None:
                                subj += self
                    else:
                        count += 1

                if count > 50:
                    break

            return csvreader



# !/usr/bin/python3
# -*- coding: utf-8 -*-

from pybpodgui_plugin.models.setup.setup_dockwindow import SetupDockWindow


class SetupCom(SetupDockWindow):
    """
    Define board actions that are triggered by setup.

    .. seealso::
        This class heavy relies on the corresponding API module.

        :py:class:`pybpodgui_api.models.setup.setup_com.SetupCom`

    **Methods**

    """

    def stop_task(self):
        """
        Stop task by calling API.

        Also, update UI by calling :py:meth:`pybpodgui_plugin.models.setup.setup_uibuisy.SetupUIBusy.update_ui`.
        """
        super(SetupCom, self).stop_task()
        self.update_ui()

    def kill_task(self):
        """
        Kill the task by calling
        """
        super(SetupCom, self).kill_task()
        self.update_ui()



# !/usr/bin/python3
# -*- coding: utf-8 -*-

import logging

from confapp import conf


import pyforms
from pyforms.basewidget import BaseWidget

from pyforms_generic_editor.models.project import Project

logger = logging.getLogger(__name__)

class ProjectsWindow(BaseWidget):
    def __init__(self, mainwindow=None):
        BaseWidget.__init__(self, 'Projects')

        self._projects = []

    def __add__(self, obj):
        self._projects.append(obj)
        return self

    def __sub__(self, obj):
        if isinstance(obj, Project): self._projects.remove(obj)
        return self

    def create_project(self):
        """
        Invoke project creation

        .. seealso::
            * Create project treenode: :class:`pybehavior.models.projects.projects_treenode.ProjectsTreeNode.create_project`.

        :rtype: Project
        """
        pass

    def open_project(self, project_path=None):
        """
        Open project

        .. seealso::
            * Open project treenode: :class:`pybehavior.models.projects.projects_treenode.ProjectsTreeNode.open_project`.

        :param str project_path:
        """
        pass

    def save_all_projects(self):
        for project in self.projects: project.save()

    def save_current_project(self):
        try:
            selected = self.tree.selected_item
            if hasattr(selected, 'window'):
                if issubclass(Project, selected.window.__class__):
                    selected.window.save()
                else:
                    selected.window.project.save()
        except Exception as err:
            logger.warning(err, exc_info=True)
            self.warning(str(err), "Error on saving project")

    def save_current_project_as(self):
        try:
            selected = self.tree.selected_item
            if hasattr(selected, 'window'):
                if issubclass(Project, selected.window.__class__):
                    selected.window.save_as()
                else:
                    selected.window.project.save_as()
        except Exception as err:
            logger.warning(err, exc_info=True)

    def is_saved(self):
        res = True
        for project in self.projects:
            if not project.is_saved():
                res = False
        return res

    @property
    def projects(self):
        return self._projects


if __name__ == "__main__": pyforms.start_app(ProjectsWindow)



# !/usr/bin/python3
# -*- coding: utf-8 -*-

import logging
import math
import socket
import sys

from confapp import conf as settings
from datetime import datetime as datetime_now
from pybpodapi.bpod.hardware.hardware import Hardware
from pybpodapi.bpod.hardware.channels import ChannelType
from pybpodapi.bpod.hardware.channels import ChannelName
from pybpodapi.bpod.hardware.events import EventName
from pybpodapi.bpod.hardware.output_channels import OutputChannel

from pybpodapi.bpod_modules.bpod_modules import BpodModules
from pybpodapi.exceptions.bpod_error import BpodErrorException

from pybpodapi.com.messaging.end_trial import EndTrial
from pybpodapi.com.messaging.trial import Trial
from pybpodapi.com.messaging.event_occurrence import EventOccurrence
from pybpodapi.com.messaging.event_resume import EventResume
from pybpodapi.com.messaging.softcode_occurrence import SoftcodeOccurrence
from pybpodapi.com.messaging.session_info import SessionInfo
from pybpodapi.com.messaging.warning import WarningMessage
from pybpodapi.com.messaging.value import ValueMessage
from pybpodapi.com.messaging.state_transition import StateTransition

from pybpodapi.session import Session

from .non_blockingstreamreader import NonBlockingStreamReader
from .non_blockingsocketreceive import NonBlockingSocketReceive

logger = logging.getLogger(__name__)


class BpodBase(object):
    """
    API to interact with Bpod

    :ivar Session session: Session for this bpod running experiment
    :ivar Hardware hardware: Hardware object representing Bpod hardware
    :ivar MessageAPI message_api: Abstracts communication with Bpod box
    :ivar bool new_sma_sent: whether a new state machine was already uploaded to Bpod box
    """

    class Events(EventName):
        pass

    class OutputChannels(OutputChannel):
        pass

    class ChannelTypes(ChannelType):
        pass

    class ChannelNames(ChannelName):
        pass

    CHECK_STATE_MACHINE_COUNTER = 0

    def __init__(self, serial_port=None, sync_channel=None, sync_mode=None, net_port=None):
        self._session = self.create_session()

        self.serial_port = serial_port if serial_port is not None else settings.PYBPOD_SERIAL_PORT
        self.baudrate = settings.PYBPOD_BAUDRATE
        self.sync_channel = sync_channel if sync_channel is not None else settings.PYBPOD_SYNC_CHANNEL
        self.sync_mode = sync_mode if sync_mode is not None else settings.PYBPOD_SYNC_MODE
        self.net_port = net_port if net_port is not None else settings.PYBPOD_NET_PORT
        self._hardware = Hardware()    # type: Hardware
        self.bpod_modules = None          # type: BpodModules
        self.bpod_start_timestamp = None

        self._new_sma_sent = False         # type: bool
        self._skip_all_trials = False

        self._hardware.sync_channel = self.sync_channel  # 255 = no sync, otherwise set to a hardware channel number
        self._hardware.sync_mode = self.sync_mode    # 0 = flip logic every trial, 1 = every state

        self.session += SessionInfo(self.session.INFO_SERIAL_PORT, self.serial_port)
        self.session += SessionInfo(self.session.INFO_PROTOCOL_NAME, settings.PYBPOD_PROTOCOL)
        self.session += SessionInfo(self.session.INFO_CREATOR_NAME, settings.PYBPOD_CREATOR)
        self.session += SessionInfo(self.session.INFO_PROJECT_NAME, settings.PYBPOD_PROJECT)
        self.session += SessionInfo(self.session.INFO_EXPERIMENT_NAME, settings.PYBPOD_EXPERIMENT)
        self.session += SessionInfo(self.session.INFO_BOARD_NAME, settings.PYBPOD_BOARD)
        self.session += SessionInfo(self.session.INFO_SETUP_NAME, settings.PYBPOD_SETUP)
        self.session += SessionInfo(self.session.INFO_BPODGUI_VERSION, settings.PYBPOD_BPODGUI_VERSION)

        if self.net_port:
            self.session += SessionInfo(self.session.INFO_NET_PORT, self.net_port)

        for subject_name in settings.PYBPOD_SUBJECTS:
            self.session += SessionInfo(self.session.INFO_SUBJECT_NAME, subject_name)

        if hasattr(settings, 'PYBPOD_VARSNAMES'):
            for varname in settings.PYBPOD_VARSNAMES:
                self.session += ValueMessage(varname, getattr(settings, varname))

    #########################################
    ############ PUBLIC METHODS #############
    #########################################

    def loop_handler(self):
        """
        handler that will execute on every loop when the bpod is running
        """
        pass

    def open(self):
        """
        Starts Bpod.

        Connect to Bpod board through serial port, test handshake, retrieve firmware version,
        retrieve hardware description, enable input ports and configure channel synchronization.

        Example:

        .. code-block:: python

            my_bpod = Bpod().open("/dev/tty.usbmodem1293", "/Users/John/Desktop/bpod_workspace", "2afc_protocol")

        :param str serial_port: serial port to connect
        :param str workspace_path: path for bpod output files (no folders will be created)
        :param str session_name: this name will be used for output files
        :param int baudrate [optional]: baudrate for serial connection
        :param int sync_channel [optional]: Serial synchronization channel: 255 = no sync, otherwise set to a hardware channel number
        :param int sync_mode [optional]: Serial synchronization mode: 0 = flip logic every trial, 1 = every state
        :return: Bpod object created
        :rtype: pybpodapi.model.bpod
        """

        logger.info("Starting Bpod")

        self._bpodcom_connect(self.serial_port, self.baudrate)

        if not self._bpodcom_handshake():
            raise BpodErrorException('Error: Bpod failed to confirm connectivity. Please reset Bpod and try again.')

        #########################################################
        ### check the firmware version ##############################
        #########################################################
        firmware_version, machine_type = self._bpodcom_firmware_version()

        if firmware_version < int(settings.TARGET_BPOD_FIRMWARE_VERSION):
            raise BpodErrorException('Error: Old firmware detected. Please update Bpod 0.7+ firmware and try again.')

        if firmware_version > int(settings.TARGET_BPOD_FIRMWARE_VERSION):
            raise BpodErrorException('Error: Future firmware detected. Please update the Bpod python software.')

        self._hardware.firmware_version = firmware_version
        self._hardware.machine_type = machine_type
        #########################################################
        self._bpodcom_hardware_description(self._hardware)

        if not self._bpodcom_enable_ports(self._hardware):
            raise BpodErrorException('Error: Failed to enable Bpod inputs.')

        if not self._bpodcom_set_sync_channel_and_mode(sync_channel=self.sync_channel, sync_mode=self.sync_mode):
            raise BpodErrorException('Error: Failed to configure syncronization.')

        # check if any module is connected
        self.bpod_modules = self._bpodcom_get_modules_info(self._hardware)

        self._hardware.setup(self.bpod_modules)

        # initialise the server to handle commands
        if self.net_port is not None:
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            self.sock.bind(('0.0.0.0', self.net_port))
            self.socketin = NonBlockingSocketReceive(self.sock)
        else:
            self.sock = None
            self.socketin = None

        # initialise the thread that will handle the stdin commands
        self.stdin = NonBlockingStreamReader(sys.stdin) if settings.PYBPOD_API_ACCEPT_STDIN else None
        #####################################################

        return self

    def close(self):
        """
        Close connection with Bpod
        """
        self.session += SessionInfo(self.session.INFO_SESSION_ENDED, datetime_now.now())

        if hasattr(settings, 'PYBPOD_VARSNAMES'):
            for varname in settings.PYBPOD_VARSNAMES:
                self.session += ValueMessage(varname, getattr(settings, varname))

        self._bpodcom_disconnect()

        del self._session

        if self.socketin is not None:
            self.socketin.close()
            self.sock.close()

        if self.stdin is not None:
            self.stdin.close()

    def stop_trial(self):
        self._bpodcom_stop_trial()

    def pause(self):
        self._bpodcom_pause_trial()

    def resume(self):
        self._bpodcom_resume_trial()

    def refresh_modules(self):
        # check if any module is connected
        self.bpod_modules = self._bpodcom_get_modules_info(self._hardware)
        self._hardware.setup(self.bpod_modules)

    def register_value(self, name, value):
        self._session += ValueMessage(name, value)

    def send_state_machine(self, sma, run_asap=None):
        """
        Builds message and sends state machine to Bpod

        :param pybpodapi.model.state_machine sma: initialized state machine
        """
        if not self.bpod_com_ready:
            raise Exception('Bpod connection is closed')

        if self._skip_all_trials is True:
            return

        logger.info("Sending state machine")

        sma.update_state_numbers()

        state_machine_body = sma.build_message() + sma.build_message_global_timer() + sma.build_message_32_bits()

        self._bpodcom_send_state_machine(sma.build_header(run_asap, len(state_machine_body)) + state_machine_body)

        self._new_sma_sent = True

    def run_state_machine(self, sma):
        """

        Adds a new trial to current session and runs state machine on Bpod box.

        While state machine is running, messages are processed accordingly.

        When state machine stops, timestamps are updated and trial events are processed.

        Finally, data is released for registered data consumers / exporters.

        .. seealso::

            Send command "run state machine": :meth:`pybpodapi.bpod.bpod_base.BpodBase.run_state_machine`.

            Process opcode: :meth:`pybpodapi.bpod.bpod_base.BpodBase._BpodBase__process_opcode`.

            Update timestamps: :meth:`pybpodapi.bpod.bpod_base.BpodBase._BpodBase__update_timestamps`.

        :param (:class:`pybpodapi.state_machine.StateMachine`) sma: initialized state machine
        """
        if not self.bpod_com_ready:
            raise Exception('Bpod connection is closed')

        if self._skip_all_trials is True:
            return False

        self.session += Trial(sma)

        logger.info("Running state machine, trial %s", len(self.session.trials))

        self.trial_timestamps = []  # Store the trial timestamps in case bpod is using live_timestamps

        self._bpodcom_run_state_machine()
        if self._new_sma_sent:
            if self._bpodcom_state_machine_installation_status():
                self._new_sma_sent = False
            else:
                raise BpodErrorException('Error: The last state machine sent was not acknowledged by the Bpod device.', self)

        self.trial_start_timestamp = self._bpodcom_get_trial_timestamp_start()

        if self.bpod_start_timestamp is None:
            self.bpod_start_timestamp = self.trial_start_timestamp

        #####################################################
        # create a list of executed states
        state_change_indexes = []

        # flags used to stop a trial (or all trials)
        interrupt_task = False
        kill_task = False

        sma.is_running = True
        while sma.is_running:

            # read commands from the stdin ######################
            if self.stdin is not None:
                inline = self.stdin.readline()
                if inline is not None:
                    interrupt_task, kill_task = self.handle_inline(inline, sma)
            #####################################################

            # read commands from a net socket ###################
            if self.socketin is not None:
                inline = self.socketin.readline()

                if inline is not None:
                    inline = inline.decode().strip()
                    interrupt_task, kill_task = self.handle_inline(inline, sma)
            #####################################################

            if self.data_available():
                opcode, data = self._bpodcom_read_opcode_message()
                self.__process_opcode(sma, opcode, data, state_change_indexes)

            self.loop_handler()

            if interrupt_task or kill_task:
                self._skip_all_trials = True
                break

        self.session += EndTrial('The trial ended')

        if not interrupt_task:
            self.__update_timestamps(sma, state_change_indexes)
            self.session.add_trial_events()

        logger.info("Publishing Bpod trial")

        if interrupt_task and kill_task:
            self.close()
            exit(0)

        return not interrupt_task

    def handle_inline(self, inline, sma):
        interrupt_task = False
        kill_task = False
        if inline.startswith('pause-trial'):
            self.pause()
        elif inline.startswith('resume-trial'):
            self.resume()
        elif inline.startswith('stop-trial'):
            self.stop_trial()
        elif inline.startswith('close'):
            self.stop_trial()
            interrupt_task = True
        elif inline.startswith('kill'):
            self.stop_trial()
            interrupt_task = kill_task = True
        elif inline.startswith('SoftCode'):
            softcode = int(inline[-1]) - 1
            self.trigger_softcode(softcode)
        elif inline.startswith('trigger_input:'):
            tdata = inline.split(':')
            chn_name = tdata[1]
            evt_data = tdata[2]
            # TODO: surround this call in a try except to capture calls with unavailable channel names
            channel_number = sma.hardware.channels.input_channel_names.index(chn_name)
            self.trigger_input(channel_number, evt_data)
        elif inline.startswith('trigger_output:'):
            tdata = inline.split(':')
            chn_name = tdata[1]
            evt_data = tdata[2]
            # TODO: surround this call in a try except to capture calls with unavailable channel names
            channel_number = sma.hardware.channels.output_channel_names.index(chn_name)
            self.trigger_output(channel_number, evt_data)
        elif inline.startswith('message:'):
            tdata = inline.split(':')
            module_index = int(tdata[1])
            msg = tdata[2]
            final_msg = []
            msg_elems = msg.split()
            if msg_elems[0].startswith('\''):
                final_msg.append(ord(msg_elems[0][1]))
            for x in msg_elems[1:]:
                final_msg.append(int(x))
            self.load_message(module_index, final_msg)

        return interrupt_task, kill_task

    def load_serial_message(self, serial_channel, message_ID, serial_message):
        """
        Load serial message on Bpod

        :param int serial_channel: Serial port to send, 1, 2 or 3
        :param int message_ID: Unique id for the message. Should be between 1 and 255
        :param list(int) serial_message: Message to send. The message should be bigger than 3 bytes.
        """
        response = self._bpodcom_load_serial_message(serial_channel, message_ID, serial_message, 1)

        if not response:
            raise BpodErrorException('Error: Failed to set serial message.')

    def reset_serial_messages(self):
        """
        Reset serial messages to equivalent byte codes (i.e. message# 4 = one byte, 0x4)
        """
        response = self._bpodcom_reset_serial_messages()

        if not response:
            raise BpodErrorException('Error: Failed to reset serial message library.')

    def softcode_handler_function(self, data):
        """
        Users can override this function directly on the protocol to handle a softcode from Bpod

        :param int data: soft code number
        """
        pass

    def echo_softcode(self, softcode):
        return self._bpodcom_echo_softcode(softcode)

    def trigger_event(self, event_index, event_data):
        return self._bpodcom_manual_override_exec_event(event_index, event_data)

    def trigger_input(self, channel_number, value):
        return self._bpodcom_override_input_state(channel_number, value)

    def trigger_output(self, channel_number, value):
        return self._bpodcom_override_digital_hardware_state(channel_number, value)

    def trigger_softcode(self, softcode):
        return self._bpodcom_send_softcode(softcode)

    def load_message(self, module_index, msg):
        # get module reference
        module = [x for x in self.modules if x.serial_port == module_index]
        # call module_write. on module reference
        if module:
            self._bpodcom_module_write(module_index - 1, msg)

    #########################################
    ############ PRIVATE METHODS ############
    #########################################

    def create_session(self):
        return Session()

    def __process_opcode(self, sma, opcode, data, state_change_indexes):
        """
        Process data from bpod board given an opcode

        In original bpod, sma.raw_data == raw_events

        :param sma: state machine object
        :param int opcode: opcode number
        :param data: data from bpod board
        :param state_change_indexes:
        :return:
        """

        current_trial = self.session.current_trial

        if opcode == 1:  # Read events
            n_current_events = data
            current_events = self._bpodcom_read_current_events(n_current_events)
            transition_event_found = False

            if self.hardware.live_timestamps:
                event_timestamp = self._bpodcom_read_event_timestamp()
            else:
                event_timestamp = None

            for event_id in current_events:
                if event_id == 255:
                    sma.is_running = False
                else:
                    self._session += EventOccurrence(
                        event_id,
                        sma.hardware.channels.get_event_name(event_id),
                        event_timestamp
                    )
                    self.trial_timestamps.append(event_timestamp)

                    # input matrix
                    if not transition_event_found:
                        logger.debug("transition event not found")
                        logger.debug("Current state: %s", sma.current_state)
                        for transition in sma.input_matrix[sma.current_state]:
                            logger.debug("Transition: %s", transition)
                            if transition[0] == event_id:
                                if sma.use_255_back_signal and transition[1] == 255:
                                    sma.current_state = current_trial.states[-2]
                                else:
                                    sma.current_state = transition[1]

                                if not math.isnan(sma.current_state):
                                    logger.debug("adding states input matrix")
                                    current_trial.states.append(sma.current_state)
                                    state_change_indexes.append(len(current_trial.events_occurrences) - 1)

                                transition_event_found = True

                    # state timer matrix
                    if not transition_event_found:
                        this_state_timer_transition = sma.state_timer_matrix[sma.current_state]
                        if event_id == sma.hardware.channels.events_positions.Tup:
                            if not (this_state_timer_transition == sma.current_state):
                                if sma.use_255_back_signal and this_state_timer_transition == 255:
                                    sma.current_state = current_trial.states[-2]
                                else:
                                    sma.current_state = this_state_timer_transition

                                if not math.isnan(sma.current_state):
                                    logger.debug("adding states state timer matrix")
                                    current_trial.states.append(sma.current_state)
                                    state_change_indexes.append(len(current_trial.events_occurrences) - 1)
                                transition_event_found = True

                    # global timers start matrix
                    if not transition_event_found:
                        for transition in sma.global_timers.start_matrix[sma.current_state]:
                            if transition[0] == event_id:
                                if sma.use_255_back_signal and transition[1] == 255:
                                    sma.current_state = current_trial.states[-2]
                                else:
                                    sma.current_state = transition[1]

                                if not math.isnan(sma.current_state):
                                    logger.debug("adding states global timers start matrix")
                                    current_trial.states.append(sma.current_state)
                                    state_change_indexes.append(len(current_trial.events_occurrences) - 1)
                                transition_event_found = True

                    # global timers end matrix
                    if not transition_event_found:
                        for transition in sma.global_timers.end_matrix[sma.current_state]:
                            if transition[0] == event_id:

                                if sma.use_255_back_signal and transition[1] == 255:
                                    sma.current_state = current_trial.states[-2]
                                else:
                                    sma.current_state = transition[1]

                                if not math.isnan(sma.current_state):
                                    logger.debug("adding states global timers end matrix")
                                    current_trial.states.append(sma.current_state)
                                    state_change_indexes.append(len(current_trial.events_occurrences) - 1)
                                transition_event_found = True

                logger.debug("States indexes: %s", current_trial.states)

            if transition_event_found and not math.isnan(sma.current_state):
                state_name = sma.state_names[sma.current_state]
                self._session += StateTransition(state_name, event_timestamp)

        elif opcode == 2:  # Handle soft code
            self._session += SoftcodeOccurrence(data)
            self.softcode_handler_function(data)

    def __update_timestamps(self, sma, state_change_indexes):
        """
        Read timestamps from Bpod and update state machine info

        :param StateMachine sma:
        :param list state_change_indexes:
        """

        current_trial = self.session.current_trial
        current_trial.trial_start_timestamp = self.trial_start_timestamp  # start timestamp of first trial
        current_trial.bpod_start_timestamp = self.bpod_start_timestamp

        trial_end_timestamp, discrepancy = self._bpodcom_read_timestamps()
        current_trial.trial_end_timestamp = trial_end_timestamp

        self.session += SessionInfo(
            self.session.INFO_TRIAL_BPODTIME,
            trial_end_timestamp-self.trial_start_timestamp,
            start_time=self.trial_start_timestamp,
            end_time=trial_end_timestamp
        )

        if discrepancy > 1:
            self.session += WarningMessage("Bpod missed hardware update deadline(s) on the past trial by ~{milliseconds}ms".format(milliseconds=discrepancy))

        if self.hardware.live_timestamps:
            timestamps = self.trial_timestamps
        else:
            timestamps = self._bpodcom_read_alltimestamps()
            timestamps = [float(t)*self._hardware.times_scale_factor for t in timestamps]

            # update the timestamps of the events #############################################################
            for event, timestamp in zip(current_trial.events_occurrences, timestamps):
                event.host_timestamp = timestamp
                e = EventResume(event.event_id, event.event_name, host_timestamp=timestamp)
                self.session += e
            ###################################################################################################

        current_trial.event_timestamps = timestamps
        current_trial.state_timestamps += [timestamps[i] for i in state_change_indexes]
        current_trial.state_timestamps += timestamps[-1:]

    def find_module_by_name(self, name):
        """
        Search for a module by name
        """
        for m in self.modules:
            if m.name == name:
                return m
        return None

    #########################################
    ############## PROPERTIES ###############
    #########################################

    @property
    def session(self):
        return self._session  # type: Session

    @session.setter
    def session(self, value):
        self._session = value  # type: Session

    @property
    def hardware(self):
        return self._hardware  # type: Hardware

    @property
    def modules(self):
        return [m for m in self.bpod_modules if m.connected]

    @property
    def skip_all_trials(self) -> bool:
        return self._skip_all_trials

    # @property
    # def inputs(self):
    #   return self._hardware.inputs

    # @property
    # def outputs(self):
    #   return self._hardware.outputs

    # @property
    # def channels(self):
    #   return self._hardware.channels

    # @property
    # def max_states(self):
    #   return self._hardware.max_states

    # @property
    # def max_serial_events(self):
    #   return self._hardware.max_serial_events

    # @property
    # def inputs_enabled(self):
    #   return self._hardware.inputs_enabled

    # @property
    # def cycle_period(self):
    #   return self._hardware.cycle_period

    # @property
    # def n_global_timers(self):
    #   return self._hardware.n_global_timers

    # @property
    # def n_global_counters(self):
    #   return self._hardware.n_global_counters

    # @property
    # def n_conditions(self):
    #   return self._hardware.n_conditions

    # @property
    # def n_uart_channels(self):
    #   return self._hardware.n_uart_channels

    # @property
    # def firmware_version(self):
    #   return self._hardware.firmware_version

    # @property
    # def machine_type(self):
    #   return self._hardware.machine_type

    # @property
    # def cycle_frequency(self):
    #   return self._hardware.cycle_frequency



# !/usr/bin/python3
# -*- coding: utf-8 -*-


import os
import datetime
import logging
from confapp import conf

logger = logging.getLogger(__name__)


class SessionBase(object):
    """
    Represents a board running session
    """

    def __init__(self, setup):
        setup += self
        self.uuid4 = None  # the session will only gain a uuid4 after the session is executed

        self.data = None
        self.setup = setup
        self.name = self.__default_name(setup)
        self.creator = ''
        self.setup_name = setup.name
        self.board_name = setup.board.name if setup.board else None
        self.task_name = setup.task.name if setup.task else None
        self.board_serial_port = setup.board.serial_port if setup.board else None
        self.started = datetime.datetime.now()
        self.ended = None
        self.messages_history = []
        self.subjects = []
        self.filepath = None
        self.variables = []
        self.user = None

    def __default_name(self, setup):
        name = conf.PYBPODGUI_API_DEFAULT_SESSION_NAME
        name = name.replace('[datetime]', datetime.datetime.now().strftime('%Y%m%d-%H%M%S'))
        name = name.replace('[project]', setup.experiment.project.name)
        name = name.replace('[experiment]', setup.experiment.name)
        name = name.replace('[setup]', setup.name)
        name = name.replace('[protocol]', setup.task.name if setup.task is not None else 'None')
        name = name.replace('[subjects]', '.'.join([s.name for s in setup.subjects]))
        return name

    ##########################################################################
    ####### PROPERTIES #######################################################
    ##########################################################################

    def __add__(self, value):
        self._messages_history.append(value)
        return self

    def remove(self):
        """
        Remove the session from the project

        """
        pass

    @property
    def setup(self):
        """
        Get and set the setup

        :rtype: Setup
        """
        return self._setup

    @setup.setter
    def setup(self, value):
        self._setup = value

    @property
    def name(self):
        """
        Get and set session name

        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def subjects(self):
        """
        Get and set session name

        :rtype:
        """
        return self._subjects

    @subjects.setter
    def subjects(self, value):
        self._subjects = value

    @property
    def path(self):
        """
        Get and set path name

        :rtype: str
        """
        if self.setup.path is None:
            return None
        return os.path.join(self.setup.path, 'sessions', self.name)

    @property
    def filepath(self):
        return self._filepath

    @filepath.setter
    def filepath(self, value):
        self._filepath = value

    @property
    def variables(self):
        return self._variables

    @variables.setter
    def variables(self, value):
        self._variables = value

    @property
    def setup_name(self):
        """
        Get and set setup name

        :rtype: str
        """
        return self._setup_name

    @setup_name.setter
    def setup_name(self, value):
        self._setup_name = value

    @property
    def board_name(self):
        """
        Get and set board name

        :rtype: str
        """
        return self._board_name

    @board_name.setter
    def board_name(self, value):
        self._board_name = value

    @property
    def task_name(self):
        """
        Get and set the task name

        :rtype: str
        """
        return self._task_name

    @task_name.setter
    def task_name(self, value):
        self._task_name = value

    @property
    def board_serial_port(self):
        """
        Get and set board serial port

        :rtype: str
        """
        return self._board_serial_port

    @board_serial_port.setter
    def board_serial_port(self, value):
        self._board_serial_port = value

    @property
    def started(self):
        """
        Get and set the start datetime of the session

        :rtype: datetime.datetime
        """
        return self._started

    @started.setter
    def started(self, value):
        self._started = value

    @property
    def ended(self):
        """
        Get and set the end datetime of the session

        :rtype: datetime.datetime
        """
        return self._ended

    @ended.setter
    def ended(self, value):
        self._ended = value

    @property
    def messages_history(self):
        """
        Get and set the history of messages

        :rtype: list(BaseMessage)
        """
        return self._messages_history

    @messages_history.setter
    def messages_history(self, value):
        self._messages_history = value

    @property
    def project(self):
        """
        Get the session Project

        :rtype: Project
        """
        return self.setup.project

    @property
    def task(self):
        """
        Get the session Task

        :rtype: Task
        """
        return self.setup.task

    @property
    def is_running(self):
        return self.status == self.STATUS_SESSION_RUNNING



# !/usr/bin/python3
# -*- coding: utf-8 -*-

import logging
from pybpodapi.com.messaging.base_message import BaseMessage
from pybpodapi.utils import date_parser

logger = logging.getLogger(__name__)


class StateTransition(BaseMessage):
    """
    Store timestamps for a specific state occurrence of the state machine

    :ivar str name: name of the state
    :ivar list(StateDuration) timestamps: a list of timestamps (start and end) that corresponds to occurrences of this state
    """

    MESSAGE_TYPE_ALIAS = "TRANSITION"
    MESSAGE_COLOR = (0, 200, 0)

    def __init__(self, state_name, host_timestamp):
        """

        :param str name: name of the state
        """
        super().__init__(state_name, host_timestamp)

    def tolist(self):
        return [
            self.MESSAGE_TYPE_ALIAS,
            str(self.pc_timestamp),
            self.host_timestamp,
            None,
            self.content,
            None,
        ]

    @classmethod
    def fromlist(cls, row):
        """
        Returns True if the typestr represents the class
        """
        obj = cls(row[4], float(row[2]) if row[2] else None)
        obj.pc_timestamp = date_parser.parse(row[1])

        return obj

    @property
    def state_name(self):
        return self.content



# !/usr/bin/python3
# -*- coding: utf-8 -*-

import logging

import pyforms as app
from pybpodgui_api.exceptions.run_setup import RunSetupError
from pybpodgui_api.models.subject import Subject
from pybpodgui_api.models.subject.subject_com import WrongSubjectConfigured
from pyforms.basewidget import BaseWidget
from pyforms.controls import ControlButton
from pyforms.controls import ControlCombo
from pyforms.controls import ControlText
from pyforms_gui.controls.control_checkbox import ControlCheckBox

logger = logging.getLogger(__name__)


class SubjectWindow(Subject, BaseWidget):
    """
    Define here which fields from the board model should appear on the details section.

    The model fields shall be defined as UI components like text fields, buttons, combo boxes, etc.

    You may also assign actions to these components.

    **Properties**

        name
            :class:`string`

            Name associated with this board. Returns the current value stored in the :py:attr:`_name` text field.

        serial_port
            :class:`string`

            Serial port associated with this board. Returns the current value stored in the :py:attr:`_serial_port` text field.


    **Private attributes**

        _name
            :class:`pyforms.controls.ControlText`

            Text field to edit board name. Editing this field fires the event :meth:`BoardWindow._BoardWindow__name_changed_evt`.

        _serial_port
            :class:`pyforms.controls.ControlText`

            Text field to edit serial port. Editing this field fires the event :meth:`BoardWindow._BoardWindow__serial_changed_evt`.

        _log_btn
            :class:`pyforms.controls.ControlButton`

            Button to show this board events on a console window. Pressing the button fires the event :class:`BoardDockWindow.open_log_window`.

        _formset
            Describe window fields organization to PyForms.

    **Methods**

    """

    def __init__(self, project=None):
        """

        :param project: project where this board belongs
        :type project: pycontrolgui.models.project.Project
        """
        BaseWidget.__init__(self, 'Subject')
        self.layout().setContentsMargins(5, 10, 5, 5)

        self._selected_setup = None

        self._name = ControlText('Name')
        self._setups = ControlCombo('Setup')
        self._run = ControlButton('Run', checkable=True, default=self._run_task)

        self._run = ControlButton('Run',
                                  checkable=True,
                                  default=self._run_task,
                                  helptext="When a task is running, you can skip all remaining trials by pressing this button. <br/> <b>NOTE:</b> This means that you will need to break the cycle in your task code when the run_state_machine method returns False.")
        self._kill_task_btn = ControlButton('Kill',
                                            default=self._kill_task,
                                            style="background-color:rgb(255,0,0);font-weight:bold;",
                                            helptext="<b>NOTE:</b>This will exit the task process abruptly. The code you might have after the trial loop won't execute.")

        self._kill_task_btn.enabled = False

        self._stoptrial_btn = ControlButton('Skip trial', default=self._stop_trial_evt)
        self._pause_btn = ControlButton('Pause', checkable=True, default=self._pause_evt)
        self._stoptrial_btn.enabled = False
        self._pause_btn.enabled = False

        self._detached = ControlCheckBox('Detach from GUI', changed_event=self.__detached_changed_evt)

        Subject.__init__(self, project)

        self._formset = [
            '_name',
            '_setups',
            '_detached',
            ('_run', '_kill_task_btn'),
            ('_stoptrial_btn', '_pause_btn'),
            ' ',
        ]

        self._name.changed_event = self.__name_changed_evt
        self.reload_setups()

    def _run_task(self):
        """
        Defines behavior of the button :attr:`SubjectWindow._run_task_btn`.

        This methods is called every time the user presses the button.
        """
        if not self.can_run_task():
            return
        try:
            if self.setup.status == self.setup.STATUS_RUNNING_TASK:
                self.setup.stop_task()
                self._run.enabled = False
            elif self.setup.status == self.setup.STATUS_READY:
                self.setup.run_task()
        except RunSetupError as err:
            self.warning(str(err), "Warning")
        except Exception as err:
            self.alert(str(err), "Unexpected Error")

    def _kill_task(self):
        """
        Kills the currently running task.
        """
        if self.setup.status == self.setup.STATUS_RUNNING_TASK:
            self.setup.kill_task()

    def _stop_trial_evt(self):
        setup = self._setups.value
        if setup:
            setup.stop_trial()
        else:
            self.critical("There isn't any setup selected. Please select one before continuing.", "No setup selected")

    def _pause_evt(self):
        self.pause_trial()

    def can_run_task(self):
        try:
            res = super().can_run_task()
        except WrongSubjectConfigured:
            if self.setup.status == self.setup.STATUS_READY:
                # warn here the user that if he/she pretends to continue, all the other subjects will be removed
                # (only if there aren't any subjects)
                result = self.question("Attention: There are subjects in the selected setup.\nDo you wish to continue?\n\n"
                                       "(note: if you continue all the existing subjects will be replaced)",
                                       "Existing subjects")
                if result == 'yes':
                    self.setup.clear_subjects()
                    res = True
                    self.setup += self
                elif result == 'no':
                    res = False
                else:
                    res = True
            else:
                res = False

        except Exception as e:
            self.warning(str(e), "Error occurred")
            res = False

        if res:
            res = self.setup.can_run_task()
        if not res:
            self._run.checked = False
        return res

    def reload_setups(self):
        tmp = self._setups.value
        self._setups.clear()
        self._setups.add_item('', None)
        for experiment in self.project.experiments:
            for setup in experiment.setups:
                self._setups.add_item(
                    "{experiment} > {setup}".format(experiment=setup.experiment.name, setup=setup.name),
                    setup
                )

        self._setups.value = tmp

    def __detached_changed_evt(self):
        setup = self._setups.value
        if setup:
            setup._detached.value = self._detached.value

    def __name_changed_evt(self):
        """
        React to changes on text field :py:attr:`_name`.

        This methods is called every time the user changes the field.
        """
        if not hasattr(self, '_update_name') or not self._update_name:
            self.name = self._name.value

    def load(self, path):
        try:
            super(Subject, self).load(path)
        except Exception as ex:
            self.warning('{first_arg}'.format(first_arg=ex.args[0]), 'Unable to load a subject')

    @property
    def name(self):
        return self._name.value

    @name.setter
    def name(self, value):
        self._update_name = True  # Flag to avoid recursive calls when editing the name text field
        self._name.value = value
        self._update_name = False

    @property
    def setup(self):
        return self._setups.value

    @setup.setter
    def setup(self, value):
        self._setups.value = value


# Execute the application
if __name__ == "__main__":
    app.start_app(SubjectWindow)



from threading import Thread, Event
from queue import Queue, Empty


class NonBlockingCSVReader:

    def __init__(self, stream):
        '''
        stream: the stream to read from.
                Usually a process' stdout or stderr.
        '''
        self._s = stream
        self._q = Queue()

        class PopulateQueue(Thread):

            def __init__(self, socket, queue):
                Thread.__init__(self)
                self.daemon = True
                self.socket = socket
                self.queue = queue
                self.event = Event()

            def run(self):
                try:
                    while True:
                        # if self.event.is_set(): break
                        line = next(stream)
                        if line:
                            self.queue.put(line)
                        else:
                            # self.event.set()
                            pass
                        # self.event.wait(0.05)
                except StopIteration:
                    pass

        self._t = PopulateQueue(self._s, self._q)
        self._t.daemon = True
        self._t.start()  # start collecting lines from the stream

    def readline(self, timeout=None):
        try:
            return self._q.get(block=timeout is not None, timeout=timeout)
        except Empty:
            return None

    def close(self):
        # self._t.event.set()
        pass


class UnexpectedEndOfStream(Exception):
    pass



import requests
import json
from confapp import conf

class Get():
    def __init__(self,_apibase):
        self.apibase = _apibase

    def allprojects(self):
        result = requests.get(conf.ALYX_PLUGIN_ADDRESS+'/projects',headers= self.apibase.headers)
        if result.ok:
            result_data = result.json()
            for a in result_data:
                print(a)
    
    def project(self,_name):
        result = requests.get(conf.ALYX_PLUGIN_ADDRESS+'/projects/'+_name,headers= self.apibase.headers)
        if result.ok:
            print(result.json())

    def alldatasettypes(self):
        result = requests.get(conf.ALYX_PLUGIN_ADDRESS+'/dataset-types',headers= self.apibase.headers)
        if result.ok:
            result_data = result.json()
            for a in result_data:
                print(a)

    def datasettype(self, _name):
        result = requests.get(conf.ALYX_PLUGIN_ADDRESS+'/dataset-types/'+_name,headers= self.apibase.headers)
        if result.ok:
            print(result.json())

    
    def alldataformats(self):
        result = requests.get(conf.ALYX_PLUGIN_ADDRESS+'/data-formats',headers= self.apibase.headers)
        if result.ok:
            result_data = result.json()
            for a in result_data:
                print(a)

    def dataformat(self, _name):
        result = requests.get(conf.ALYX_PLUGIN_ADDRESS+'/data-formats/'+_name,headers= self.apibase.headers)
        if result.ok:
            print(result.json())

    def alldatasets(self):
        result = requests.get(conf.ALYX_PLUGIN_ADDRESS+'/datasets',headers= self.apibase.headers)
        if result.ok:
            result_data = result.json()
            for a in result_data:
                print(a)
    
    def dataset(self,_name):
        result = requests.get(conf.ALYX_PLUGIN_ADDRESS+'/datasets/'+_name,headers= self.apibase.headers)
        if result.ok:
            print(result.json())

    def allfiles(self):
        result = requests.get(conf.ALYX_PLUGIN_ADDRESS+'/files',headers= self.apibase.headers)
        if result.ok:
            result_data = result.json()
            for a in result_data:
                print(a)
    
    def files(self,_name):
        result = requests.get(conf.ALYX_PLUGIN_ADDRESS+'/files/'+_name,headers= self.apibase.headers)
        if result.ok:
            print(result.json())


    def alldatarepositorytypes(self):
        result = requests.get(conf.ALYX_PLUGIN_ADDRESS+'/data-repository-type',headers= self.apibase.headers)
        if result.ok:
            result_data = result.json()
            for a in result_data:
                print(a)

    def repositorytype(self, _name):
        result = requests.get(conf.ALYX_PLUGIN_ADDRESS+'/data-repository-type/'+_name,headers= self.apibase.headers)
        if result.ok:
            print(result.json())

    def alldatarepositories(self):
        result = requests.get(conf.ALYX_PLUGIN_ADDRESS+'/data-repository',headers= self.apibase.headers)
        if result.ok:
            result_data = result.json()
            for a in result_data:
                print(a)

    def datarepositoryty(self, _name):
        result = requests.get(conf.ALYX_PLUGIN_ADDRESS+'/data-repository/'+_name,headers= self.apibase.headers)
        if result.ok:
            print(result.json())




# !/usr/bin/python3
# -*- coding: utf-8 -*-

from confapp import conf

from pybpodgui_plugin.models.task.task_dockwindow import TaskDockWindow

Task = type(
    'Task',
    tuple(conf.GENERIC_EDITOR_PLUGINS_FINDER.find_class('models.task.Task') + [TaskDockWindow]),
    {}
)



# !/usr/bin/python3
# -*- coding: utf-8 -*-
from pybpodapi.com.messaging.base_message import BaseMessage
from pybpodapi.utils import date_parser


class EventOccurrence(BaseMessage):
    """
    Message from board that represents state change (an event)

    :ivar str event_name: name of the event
    :ivar int event_id: index of the event
    :ivar float board_timestamp: timestamp associated with this event (from bpod)

    """

    MESSAGE_TYPE_ALIAS = "EVENT"

    def __init__(self, event_id, event_name, host_timestamp=None):
        """

        :param event_id:
        :param event_name:
        :param host_timestamp:
        """
        super(EventOccurrence, self).__init__(event_name, host_timestamp)
        self._event_id = event_id

    @property
    def event_name(self):
        return self.content

    @property
    def event_id(self):
        return self._event_id

    def tolist(self):
        return [
            self.MESSAGE_TYPE_ALIAS,
            str(self.pc_timestamp),
            self.host_timestamp,
            None,
            self.event_id,
            self.event_name,
        ]

    @classmethod
    def fromlist(cls, row):
        """
        Returns True if the typestr represents the class
        """
        obj = cls(int(row[4]), row[5], float(row[2]) if row[2] else None)
        obj.pc_timestamp = date_parser.parse(row[1])

        return obj



# !/usr/bin/python3
# -*- coding: utf-8 -*-

import logging
import os

from AnyQt.QtWidgets import QFileDialog, QMessageBox

import pyforms as app
from pyforms.controls import ControlText

from pyforms_generic_editor.models.project import GenericProject

from pybpodgui_api.models.project import Project

logger = logging.getLogger(__name__)


class ProjectWindow(Project, GenericProject):
    """ ProjectWindow represents the project entity as a GUI window"""

    def __init__(self):
        """

        """

        GenericProject.__init__(self)

        self._name = ControlText('Project name')

        self.formset = ['_name', ' ']

        self._name.changed_event = self._name_changed_evt

        Project.__init__(self)

    def _name_changed_evt(self):
        if not hasattr(self, '_update_name') or not self._update_name:
            self.name = self._name.value

    @property
    def name(self):
        return self._name.value

    @name.setter
    def name(self, value):
        self._update_name = True  # Flag to avoid recurse calls when editing the name text field
        self._name.value = value
        self._update_name = False

    def save(self, project_path=None):
        if project_path:
            Project.save(self, project_path)
        elif self.path:
            Project.save(self, self.path)
        else:
            folder = QFileDialog.getExistingDirectory(self, "Select a directory to save the project: {0}".format(
                self.name))
            if folder:
                folder = os.path.join(folder, self.name)
                try:
                    Project.save(self, str(folder))
                except FileExistsError as err:
                    logger.warning(str(err))
                    QMessageBox.warning(self, 'Project exists',
                                        'Project with same name already exists. Please select another path.')

    def close(self, silent=False):
        self.projects -= self
        super(ProjectWindow, self).close(silent)


# Execute the application
if __name__ == "__main__":
    app.startApp(ProjectWindow)



# !/usr/bin/python3
# -*- coding: utf-8 -*-

import pyforms
from confapp import conf

from AnyQt.QtWidgets import QFileDialog
from AnyQt.QtCore import pyqtSignal

from pybpodgui_plugin.models.project import Project
from pybpodgui_plugin.models.session import Session

from pyforms.controls import ControlToolButton


class ProjectsWindow(object):
    """
    See:
     - pyforms_generic_editor.models.projects.__init__.py
     - pyforms_generic_editor.models.projects.projects_window.py
    """

    signal_session_create_treenode = pyqtSignal(Session)

    def __init__(self, mainwindow=None):
        super(ProjectsWindow, self).__init__(mainwindow)

    def register_on_toolbar(self, toolbar):
        toolbar += [
            ControlToolButton(
                'New',
                icon=conf.NEW_SMALL_ICON,
                default=self.create_project,
                maxheight=30
            ),
            ControlToolButton(
                'Open',
                icon=conf.OPEN_SMALL_ICON,
                default=self._option_open_project,
                maxheight=30
            ),
            ControlToolButton(
                'Save',
                icon=conf.SAVE_SMALL_ICON,
                default=self.save_current_project,
                maxheight=30
            ), '|'
        ]

    def create_project(self):
        """
        Invoke project creation

        .. seealso::
            * Create project treenode: :class:`pybpodgui_plugin.models.projects.projects_treenode.ProjectsTreeNode.create_project`.

        :rtype: Project
        """
        return Project(self)

    def open_project(self, project_path=None):
        """
        Open project

        .. seealso::
            * Open project treenode: :class:`pybpodgui_plugin.models.projects.projects_treenode.ProjectsTreeNode.open_project`.

        :param str project_path:
        """
        project = None
        if not project_path:
            project_path = QFileDialog.getExistingDirectory(self, "Select the project directory")
        if project_path:
            project = self.create_project()
            project.load(str(project_path))

        return project


if __name__ == "__main__":
    pyforms.start_app(ProjectsWindow)



# !/usr/bin/python3
# -*- coding: utf-8 -*-

import logging

import pyforms as app

from pyforms.basewidget import BaseWidget
from pyforms.controls import ControlText
from pyforms.controls import ControlList
from pyforms.controls import ControlButton
from pyforms.controls import ControlCombo
from pyforms.controls import ControlCheckBox
from pyforms.controls import ControlCheckBoxList
from pyforms.controls import ControlEmptyWidget

from pybpodgui_plugin.models.subject import Subject
from pybpodgui_api.models.setup import Setup
from pybpodgui_api.exceptions.run_setup import RunSetupError

from pybpodgui_plugin.models.setup.board_task import BoardTask
from pybpodgui_plugin.models.session import Session

logger = logging.getLogger(__name__)


class SubjectSelectPopup(BaseWidget):

    def __init__(self, subjectlist, selectedsubjects):

        super(SubjectSelectPopup, self).__init__('Add Subjects')

        self._ok_btn = ControlButton('OK')
        self._cancel_btn = ControlButton('Cancel', default=self.cancel_evt)
        self._subjectslist = ControlCheckBoxList('Subject List')

        self._formset = [
            '',
            ('', '_subjectslist', ''),
            ('', '_ok_btn', '_cancel_btn', ''),
            ''
        ]

        for subject in sorted([s for s in subjectlist], key=lambda x: x.name.lower()):
            b = False
            for a in selectedsubjects.value:
                if subject.name == a[0]:
                    self._subjectslist += (subject, True)
                    b = True
            if b is False:
                self._subjectslist += (subject, False)

    def ok_evt(self):
        self.close()
        # self.ok_event(self._subjectslist.value)

    def closewidnow(self):
        self.close()

    def subjectlist(self):
        return self._subjectslist.items

    # THIS IS THE WAY WE CREATE A SIGNAL INSIDE PYBPOD (SCROLL DOWN...)
    def ok_event(self, subjects):
        pass

    def cancel_evt(self):
        print('cancel')
        self.close()


class SetupWindow(Setup, BaseWidget):
    """
    Define here which fields from the setup model should appear on the details section.

    The model fields shall be defined as UI components like text fields, buttons, combo boxes, etc.

    You may also assign actions to these components.

    **Properties**

        name
            :class:`string`

            Name associated with this setup. Returns the current value stored in the :py:attr:`_name` text field.

        board
            :class:`pybpodgui_plugin.models.board.board_dockwindow.BoardDockWindow`

            Board associated with this setup. Returns the current value stored in the :py:attr:`_board` combo box.

    **Private attributes**

        _name
            :class:`pyforms.controls.ControlText`

            Text field to edit board name. Editing this field fires the event :meth:`SetupWindow._SetupWindow__name_changed_evt`.

        _board
            :class:`pyforms.controls.ControlCombo`

            Combo box to select board associated with this setup. Editing this field fires the event :meth:`SetupWindow._SetupWindow__board_changed_evt`.

        _run_task_btn
            :class:`pyforms.controls.ControlButton`

            Button to run task on board. Pressing the button fires the event :meth:`SetupWindow._SetupWindow__run_task`.

        _formset
            Describe window fields organization to PyForms.

    **Methods**

    """

    def __init__(self, experiment=None):
        """

        :param experiment: Experiment this setup belongs to.
        """
        BaseWidget.__init__(self, 'Experiment')
        self.layout().setContentsMargins(5, 10, 5, 5)

        self._name = ControlText('Setup name')
        self._board = ControlCombo('Board')

        self._stoptrial_btn = ControlButton('Skip trial', default=self._stop_trial_evt)
        self._pause_btn = ControlButton('Pause', checkable=True, default=self._pause_evt)
        self._run_task_btn = ControlButton('Run',
                                           checkable=True,
                                           default=self._run_task,
                                           helptext="When a task is running, you can stop all remaining trials by pressing this button. <br/> <b>NOTE:</b> This means that you will need to break the cycle yourself in your task code when the run_state_machine method returns False.")
        self._kill_task_btn = ControlButton('Kill',
                                            default=self._kill_task,
                                            style="background-color:rgb(255,0,0);font-weight:bold;",
                                            helptext="<b>NOTE:</b>This will exit the task process abruptly. The code you might have after the trial loop won't execute.")

        self._subjects_list = ControlList('Subjects', remove_function=self.__remove_subject)
        self._add_subject = ControlButton('Add subject')
        self._allsubjects = ControlCombo('Add subject')
        self._task = ControlCombo('Protocol', changed_event=self._task_changed_evt)

        self._detached = ControlCheckBox('Detach from GUI')

        self._varspanel = ControlEmptyWidget()
        self._btn = ControlButton('Open')

        Setup.__init__(self, experiment)

        self.reload_setups()
        self.reload_boards()
        self.reload_tasks()

        self._formset = [
            '_name',
            '_board',
            '_task',
            '_detached',
            ('_run_task_btn', '_kill_task_btn'),
            ('_stoptrial_btn', '_pause_btn'),
            #' ',
            {
                'Subjects': [
                    # '_allsubjects',
                    '',
                    '_add_subject',
                    '_subjects_list',
                ],
                'Variables':[
                    '_varspanel',
                ],
            }
        ]

        self._kill_task_btn.enabled = False
        self._subjects_list.readonly = True
        self._varspanel.value = self.board_task
        self._add_subject.value = self.__add_subject
        self._name.changed_event = self.__name_changed_evt
        self._board.changed_event = self._board_changed_evt

    def slot(self):
        self.clear_subjects()
        listedsubjects = self.sswindow.subjectlist()
        for subj in listedsubjects:
            if subj[1] is True:
                self += subj[0]
        self.sswindow.closewidnow()

    def reload_tasks(self, current_selected_task=None):
        # type: (current_selected_task) -> None
        """
        Reload tasks now

        :param current_selected_task: current selected task
        :type current_selected_task: pybpodgui_plugin.models.task.Task
        """
        self._task.clear()
        self._task.add_item('', 0)
        for task in self.project.tasks:
            self._task.add_item(task.name, task)
        self._task.current_index = 0
        if current_selected_task:
            self.task = current_selected_task

    def _task_changed_evt(self):
        if hasattr(self, '_update_task'):
            return
        self.task = self._task.value

    def __add__(self, obj):
        res = super(SetupWindow, self).__add__(obj)
        if isinstance(obj, Subject):
            self._subjects_list.value = [[s.name] for s in self.subjects]
        return res

    def __sub__(self, obj):
        res = super(SetupWindow, self).__sub__(obj)
        if isinstance(obj, Subject):
            self._subjects_list.value = [[s.name] for s in self.subjects]
        return res

    def __open_subject_select(self):
        self.sswindow = SubjectSelectPopup(self.project.subjects, self._subjects_list)
        self.sswindow._ok_btn.value = self.slot
        self.sswindow.show()

    def __add_subject(self):
        self.__open_subject_select()

    def __remove_subject(self):
        if self._subjects_list.selected_row_index is not None:
            name = self._subjects_list.value[self._subjects_list.selected_row_index][0]
            subject = self.project.find_subject(name)
            self -= subject
            self._subjects_list -= -1

    def _stop_trial_evt(self):
        self.stop_trial()

    def _pause_evt(self):
        if self._pause_btn.checked:
            self.pause_trial()
        else:
            self.resume_trial()

    def can_run_task(self):
        try:
            return super().can_run_task()
        except Exception as err:
            self.alert(str(err), "Unexpected Error")
            self._run_task_btn.checked = False
            return False

    def _run_task(self):
        """
        Defines behavior of the button :attr:`SetupWindow._run_task_btn`.

        This methods is called every time the user presses the button.
        """
        try:
            if self.status == SetupWindow.STATUS_RUNNING_TASK:
                self.stop_task()
            elif self.status == SetupWindow.STATUS_READY:
                self.run_task()
        except RunSetupError as err:
            self.warning(str(err), "Warning")
        except Exception as err:
            self.alert(str(err), "Unexpected Error")

    def _kill_task(self):
        """
        Kills a running task. This will stop the current trial and exit the task abruptly within the trial loop (if any).
        """
        if self.status == SetupWindow.STATUS_RUNNING_TASK:
            self.kill_task()

    def _board_changed_evt(self):
        """
        React to changes on text field :py:attr:`_board`.

        This method is called every time the user changes the field and forces a UI refresh.
        """
        if hasattr(self, '_update_board'):
            return
        self.board = self._board.value
        self.update_ui()

    def __name_changed_evt(self):
        """
        React to changes on text field :py:attr:`_name`.

        This methods is called every time the user changes the field.
        """
        if not hasattr(self, '_update_name') or not self._update_name:
            self.name = self._name.value
            self.reload_setups()

    def reload_setups(self):
        for subject in self.project.subjects:
            subject.reload_setups()

    def reload_boards(self, current_selected_board=None):
        """
        Reload boards list on combo box

        This method is fired by:
            * setup creation: :py:meth:`pybpodgui_plugin.models.setup.setup_window.SetupWindow._SetupWindow__init__`.
            * setup details section focus (dockwindow): :py:meth:`pybpodgui_plugin.models.setup.setup_dockwindow.SetupDockWindow.show`.

        :param current_selected_board: optional specify current selected board to restore after list update
        """
        self._board.clear()
        self._board.add_item('', 0)
        for board in self.project.boards:
            self._board.add_item(board.name, board)
        self._board.current_index = 0

        if current_selected_board:
            self.board = current_selected_board

        self._allsubjects.clear()
        self._allsubjects.add_item('', 0)
        for subject in sorted([s for s in self.project.subjects], key=lambda x: x.name.lower()):
            self._allsubjects.add_item(subject.name, subject)
        self._allsubjects.current_index = 0

        self._subjects_list.value = [[s.name] for s in self.subjects]

    def create_board_task(self):
        """
        Creates a new board task by calling the API.

        .. seealso::
            :py:class:`pybpodgui_api.models.setup.board_task.BoardTask`.
        """
        return BoardTask(self)

    def create_session(self):
        """
        Creates a new session by calling the API.

        .. seealso::
            :py:class:`pybpodgui_api.models.session.session_base.SessionBase`.
        """
        return Session(self)

    @property
    def name(self):
        return self._name.value

    @name.setter
    def name(self, value):
        self._update_name = True  # Flag to avoid recurse calls when editing the name text field
        self._name.value = value
        self._update_name = False
        # Update the session windows names
        if hasattr(self, 'sessions'):
            for session in self.sessions:
                session.name = session.name

    @property
    def board(self):
        if isinstance(self._board.value, str) or isinstance(self._board.value, int): return None
        return self._board.value

    @board.setter
    def board(self, value):
        if isinstance(value, str):
            value = self.project.find_board(value)
        self._update_board = True  # Flag to avoid recurse calls when editing the name text field

        if value not in self._board.values:
            self.reload_boards()
        self._board.value = value
        del self._update_board
        Setup.board.fset(self, value)

    @property
    def task(self):
        if isinstance(self._task.value, str) or isinstance(self._task.value, int):
            return None
        return self._task.value

    @task.setter
    def task(self, value):
        if isinstance(value, str):
            value = self.project.find_task(value)

        self._update_task = True  # Flag to avoid recurse calls when editing the name text field

        if value not in self._task.values:
            self.reload_tasks()

        self._task.value = value
        del self._update_task
        Setup.task.fset(self, value)

    @property
    def detached(self): return self._detached.value
    @detached.setter
    def detached(self, value): self._detached.value = value


# Execute the application
if __name__ == "__main__":
    app.start_app(SetupWindow)



#!/usr/bin/python3
# -*- coding: utf-8 -*-

import logging
import shutil
import pyforms
import os
from pyforms_gui.organizers import vsplitter
from pyforms.basewidget import BaseWidget
from pyforms.controls import ControlTreeView
from pyforms.controls import ControlCodeEditor

from AnyQt.QtCore import Qt
from AnyQt.QtWidgets import QFileSystemModel

logger = logging.getLogger(__name__)


class CodeEditor(BaseWidget):

    def __init__(self, task):
        BaseWidget.__init__(self, task.name if task else '')
        self.set_margin(5)
        self.task = task

        self._taskselected = False
        self._fileselected = None

        root_path = os.path.abspath(task.path)
        syspath_model = QFileSystemModel(self)
        syspath_model.setRootPath(root_path)
        syspath_model.setNameFilters(['*.py'])
        syspath_model.setNameFilterDisables(False)
        self.syspath_model = syspath_model

        self._browser = ControlTreeView('Files browser', default=syspath_model)
        self._code = ControlCodeEditor(
            changed_event=self.__code_changed_evt,
            discard_event=self.__code_discard_evt
        )
        self._browser.setSortingEnabled(True)

        self.formset = [vsplitter('_browser', '||', '_code')]

        for i in range(1, 4):
            self._browser.hideColumn(i)
        self._browser.item_selection_changed_event = self.__item_selection_changed_evt

        self.refresh_directory()
        self.select_file(task.filepath)

        self._browser.add_popup_menu_option('New module', self.__create_module_evt)
        self._browser.add_popup_menu_option('New module folder', self.__create_submodule_evt)
        self._browser.add_popup_menu_option('-')
        self._browser.add_popup_menu_option('Rename', self.__rename_evt)
        self._browser.add_popup_menu_option('Delete', self.__delete_evt)

    def __load_file_content(self, filepath):
        try:
            with open(filepath, "r") as file:
                self._code.value = file.read()
                self._code.enabled = True
                self._taskselected = (filepath == os.path.abspath(self.task.filepath))
                self._fileselected = os.path.relpath(filepath, self.task.path)
        except:
            self._code.value = ''
            self._code.enabled = False
            self._taskselected = False
            self._fileselected = None

    def select_file(self, filepath):
        f = self._browser.value.index(os.path.abspath(filepath))
        self._browser.setCurrentIndex(f)
        self._browser.value.sort(0, Qt.AscendingOrder)
        self._browser.sortByColumn(0, Qt.AscendingOrder)

    def __dummy(self, x, y):
        return

    def refresh_directory(self):
        root_index = self._browser.value.index(self.task.path)
        self._browser.setRootIndex(root_index)

        if self.selected_file:
            self._browser.item_selection_changed_event = self.__dummy
            self.select_file(self.selected_file)
            self._browser.item_selection_changed_event = self.__item_selection_changed_evt

        self._browser.value.sort(0, Qt.AscendingOrder)
        self._browser.sortByColumn(0, Qt.AscendingOrder)

    def __create_module_evt(self):
        name = self.input_text('Enter the module name', 'Module name')
        if name:
            folder_path = None
            for index in self._browser.selectedIndexes():
                folder_path = self.syspath_model.filePath(index)
                break
            if folder_path is not None:
                path = os.path.dirname(folder_path) if os.path.isfile(folder_path) else folder_path
            else:
                path = self.task.path

            new_filename = os.path.join(path, name+'.py')
            if not os.path.isfile(new_filename):
                with open(new_filename, "w") as file:
                    pass

    def __create_submodule_evt(self):
        name = self.input_text('Enter the module folder name', 'Module folder name')
        if name:
            
            folder_path = None
            for index in self._browser.selectedIndexes():
                folder_path = self.syspath_model.filePath(index)
                break
            if folder_path is not None:
                path = os.path.dirname(folder_path) if os.path.isfile(folder_path) else folder_path
            else:
                path = self.task.path

            try:
                new_folder = os.path.join(path, name)
                os.makedirs(new_folder)
                new_filename = os.path.join(new_folder, '__init__.py')
                with open(new_filename, "w") as file:
                    pass
            except:
                pass

    def __rename_evt(self):
        fullpath = None
        for index in self._browser.selectedIndexes():
            fullpath = self.syspath_model.filePath(index)
            break

        if os.path.isfile(fullpath):
            name, extension = os.path.splitext(os.path.basename(fullpath))
        else:
            name = os.path.basename(fullpath)

        if name == self.task.name:
            self.warning('Please use the name field in the task details window to rename the task')
            return

        name = self.input_text('Enter the new name', 'New name', name)
        if name:
            if os.path.isfile(fullpath):
                new_path = os.path.join(os.path.dirname(fullpath), name+extension)
            else:
                new_path = os.path.join(os.path.dirname(fullpath), name)
            os.rename(fullpath, new_path)

    def __delete_evt(self):
        fullpath = None
        for index in self._browser.selectedIndexes():
            fullpath = self.syspath_model.filePath(index)
            break
        reply = self.question('Please confirm the module [{0}] is to be deleted?'.format(os.path.basename(fullpath)))
        if reply == 'yes':
            if os.path.isfile(fullpath):
                os.remove(fullpath)
            else:
                shutil.rmtree(fullpath)

    def __item_selection_changed_evt(self, selected, deselected):

        if self._code.is_modified:
            self._browser.selectionModel().reset()
            self.info(
                'Please save or discard your modifications first.',
                'There are still uncommitted modifications'
            )
            if self.selected_file:
                self._browser.item_selection_changed_event = self.__dummy
                self.select_file(self.selected_file)
                self._browser.item_selection_changed_event = self.__item_selection_changed_evt
        else:
            filepath = None
            for index in self._browser.selectedIndexes():
                filepath = self.syspath_model.filePath(index)
                break
            if filepath is not None:
                self.__load_file_content(filepath)

    def __code_changed_evt(self):
        filepath = self.selected_file

        if filepath is None:
            self.warning('It is not possible to save the file',
                         'The project does not exists yet. Please save it before to be able save this file.')

        # in case the file task file not exists yet
        if not self.task.filepath or not os.path.exists(self.task.filepath):

            # check if the tasks path exists, if not create it
            tasks_path = os.path.join(self.task.project.path, 'tasks')
            if not os.path.exists(tasks_path):
                os.makedirs(tasks_path)

            # check if the task path exists, if not create it
            task_folder = os.path.join(tasks_path, self.task.name)
            if not os.path.exists(task_folder):
                os.makedirs(task_folder)

            # create an empty __init__.py file
            initfile = os.path.join(task_folder, '__init__.py')
            if not os.path.exists(initfile):
                with open(initfile, "w") as file:
                    pass

        # save the code to the file
        with open(filepath, "w") as file:
            file.write(self._code.value)

        return True

    def __code_discard_evt(self):
        filepath = None
        for index in self._browser.selectedIndexes():
            filepath = self.syspath_model.filePath(index)
            break
        if filepath is not None:
            self.__load_file_content(filepath)

        return True

    @property
    def selected_file(self):
        if self._taskselected:
            return self.task.filepath
        elif self._fileselected:
            return os.path.join(self.task.path, self._fileselected)
        else:
            return None

    def beforeClose(self):
        """
        Before closing window, ask user if she wants to save (if there are changes)

        .. seealso::
            :py:meth:`pyforms.gui.Controls.ControlMdiArea.ControlMdiArea._subWindowClosed`.

        """
        if self._code.is_modified:
            reply = self.question('Save the changes', 'Save the file')

            if reply == 'yes':
                self.__code_changed_evt()

    @property
    def title(self):
        return BaseWidget.title.fget(self)

    @title.setter
    def title(self, value):
        BaseWidget.title.fset(self, "{0} task editor".format(value))


# Execute the application
if __name__ == "__main__":
    pyforms.start_app(CodeEditor)



from pybpodapi.com.arcom import ArCOM, ArduinoTypes


class RotaryEncoderModule(object):

    COM_HANDSHAKE        = 'C'
    COM_TOGGLEEVTTRANSM  = ord('V')
    COM_TOGGLEOUTPUTSTREAM = ord('O')
    COM_TOGGLESTREAM     = ord('S')
    COM_STARTLOGGING     = ord('L')
    COM_STOPLOGGING      = ord('F')
    COM_GETLOGDATA       = ord('R')
    COM_GETCURRENTPOS    = ord('Q')
    COM_SETZEROPOS       = ord('Z')
    COM_SETPOS           = ord('P')
    COM_ENABLETHRESHOLDS = ord(';')
    COM_SETPREFIX        = ord('I')
    COM_SETTHRESHOLDS    = ord('T')
    COM_SETWRAPPOINT     = ord('W')

    def __init__(self, serialport=None):
        """
        Constructor of the RotaryEncoderModule object
        A serial connection to the Rotary Encoder board is opened at the construction of the object.

        :ivar str serialport: PC serial port where the module is connected
        """
        if serialport:
            self.open(serialport)

    def open(self, serialport):
        """
        Opens a serial connection to the Rotary Encoder board.

        :ivar str serialport: PC serial port where the module is connected
        """
        self.arcom = ArCOM().open(serialport, 115200)
        self.arcom.write_char(self.COM_HANDSHAKE)
        if self.arcom.read_uint8() != 217:
            raise Exception('Could not connect =( ')

    def close(self):
        """
        Closes the serial connection to the Rotary Encoder board.
        """
        self.arcom.close()

    def __pos_2_degrees(self, pos):
        return round(((float(pos)/512.0)*180.0)*10.0)/10.0

    def __degrees_2_pos(self, degrees):
        return int(round((float(degrees)/180.0)*512.0, 0))

    def enable_evt_transmission(self):
        """
        Enables the transmission of threshold crossing events to the Bpod state machine.
        """
        self.arcom.write_array([self.COM_TOGGLEEVTTRANSM, 1])
        return self.arcom.read_uint8() == 1

    def disable_evt_transmission(self):
        """
        Disables the transmission of events.
        """
        self.arcom.write_array([self.COM_TOGGLEEVTTRANSM, 0])
        return self.arcom.read_uint8() == 1

    def enable_module_outputstream(self):
        """
        Enables the streaming of current position data directly to another Bpod module (e.g. DDS, AnalogOutput).
        """
        self.arcom.write_array([self.COM_TOGGLEOUTPUTSTREAM, 1])
        return self.arcom.read_uint8() == 1

    def disable_module_outputstream(self):
        """
        Disables the streaming of current position data directly to another Bpod module.
        """
        self.arcom.write_array([self.COM_TOGGLEOUTPUTSTREAM, 0])
        return self.arcom.read_uint8() == 1

    def enable_stream(self):
        """
        Enables the streaming of the position and the time measurements to the USB port.
        """
        self.arcom.write_array([self.COM_TOGGLESTREAM, 1])

    def disable_stream(self):
        """
        Disables the streaming of the position and the time measurements to the USB port.
        """
        self.arcom.write_array([self.COM_TOGGLESTREAM, 0])

    def read_stream(self):
        """
        Reads the data being streamed through the USB port.
        """
        res = []
        available = self.arcom.bytes_available()

        if available > 1:
            msg = self.arcom.read_bytes_array(available)

            while len(msg) >= 7:
                if msg[0] == b'P':
                    position = int.from_bytes(b''.join(msg[1:3]), byteorder='little', signed=True)
                    evt_time = float(int.from_bytes(b''.join(msg[3:7]), byteorder='little', signed=False))/1000.0
                    position_degrees = self.__pos_2_degrees(position)
                    res.append(['P', evt_time, position_degrees])
                elif msg[0] == b'E':
                    origin = msg[1]
                    event = msg[2]
                    evt_time = float(int.from_bytes(b''.join(msg[3:7]), byteorder='little', signed=False))/1000.0
                    res.append(['E', evt_time, origin, event])

                msg = msg[7:]

        return res

    def current_position(self):
        """
        Retrieves the current position.
        """
        self.arcom.write_array([self.COM_GETCURRENTPOS])
        data_in_bytes = b''.join(self.arcom.read_bytes_array(2))
        ticks = int.from_bytes(data_in_bytes, byteorder='little', signed=True)
        return self.__pos_2_degrees(ticks)

    def set_zero_position(self):
        """
        Sets current rotary encoder position to zero.
        """
        self.arcom.write_array([self.COM_SETZEROPOS])

    def set_position(self, degrees):
        """
        Sets the current position in degrees.

        :ivar int degrees: current position in degrees.
        """
        ticks = self.__degrees_2_pos(degrees)
        data = ArduinoTypes.get_uint8_array([self.COM_SETPOS])
        data += ticks.to_bytes(2, byteorder='little', signed=True)

        self.arcom.write_array(data)
        return self.arcom.read_uint8() == 1

    def enable_thresholds(self, thresholds):
        """
        Enables the thresholds.

        :ivar list(boolean) thresholds: list of 6 booleans indicating which thresholds are active to trigger events.
        """
        if len(thresholds) != 8:
            raise Exception('Thresholds array has to be of length 8')

        string = ''.join(map(lambda x: str(int(x)), thresholds))
        bits = int(string, 2)
        self.arcom.write_array([self.COM_ENABLETHRESHOLDS, bits])

    def enable_logging(self):
        """
        Enables the logging to the SD Card.
        """
        self.arcom.write_array([self.COM_STARTLOGGING])

    def disable_logging(self):
        """
        Disables the logging to the SD Card.
        """
        self.arcom.write_array([self.COM_STOPLOGGING])

    def get_logged_data(self):
        """
        Retrieves the logged data in the SD Card.
        """
        self.arcom.write_array([self.COM_GETLOGDATA])
        msg = self.arcom.read_bytes_array(4)
        n_logs = int.from_bytes(b''.join(msg), byteorder='little', signed=False)
        data = []

        for i in range(0, n_logs):
            msg = self.arcom.read_bytes_array(8)
            data_in_bytes = b''.join(msg)

            position = int.from_bytes(data_in_bytes[:4], byteorder='little', signed=True)
            evt_time = float(int.from_bytes(data_in_bytes[4:], byteorder='little', signed=False))/1000.0
            position_degrees = self.__pos_2_degrees(position)
            data.append((evt_time, position_degrees))

        return data

    def set_prefix(self, prefix):
        """
        Sets 1-character prefix for module output stream.

        :ivar char prefix: One character to be used as prefix.
        """
        self.arcom.write_array([self.COM_SETPREFIX, prefix])
        return self.arcom.read_uint8() == 1

    def set_thresholds(self, thresholds):
        """
        Sets the thresholds values to trigger the events.

        :ivar list(int) thresholds: List, in maximum, of 6 thresholds to trigger events.
        """
        data = ArduinoTypes.get_uint8_array([self.COM_SETTHRESHOLDS, len(thresholds)])
        data += ArduinoTypes.get_uint16_array([self.__degrees_2_pos(thresh) for thresh in thresholds])
        self.arcom.write_array(data)
        return self.arcom.read_uint8() == 1

    def set_wrappoint(self, wrap_point):
        """
        Sets wrap point (number of tics in a half-rotation)

        :ivar int wrap_point: number of tics in a half-rotation.
        """
        ticks = self.__degrees_2_pos(wrap_point)
        self.arcom.write_array([self.COM_SETWRAPPOINT] + ArduinoTypes.get_uint16_array([ticks]))
        return self.arcom.read_uint8() == 1


if __name__ == '__main__':
    import time

    m = RotaryEncoderModule('/dev/ttyACM0')

    m.enable_logging()
    time.sleep(5)
    m.disable_logging()

    print(m.get_logged_data())

    """
    m.enable_stream()



    count = 0
    while count<100 or True:
        data = m.read_stream()
        if len(data)==0: continue

        print(data)
        count += 1

    m.disable_stream()

    print('set', m.set_position(179))
    m.set_zero_position()

    m.enable_thresholds([True, False, True, True, False, False, True, True])
    print(m.current_position())
    print(m.get_logged_data())
    """
    m.close()



# !/usr/bin/python3
# -*- coding: utf-8 -*-
import logging

logger = logging.getLogger(__name__)


class TaskVariable(object):
    def __init__(self, board_task, name=None, value=None, datatype='string'):
        self.board_task = board_task
        self.name = name
        self.value = value
        self.datatype = datatype

        board_task += self

    def __str__(self):
        return "{name} = {value}".format(
            name=self.name,
            value=("'"+self.value+"'") if self.datatype == 'string' else self.value
        )

    @property
    def name(self):
        """
        Get and set variable name

        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def datatype(self):
        """
        Get and set variable data type. It can be from type 'number' or 'string'

        :rtype: str
        """
        return self._datatype

    @datatype.setter
    def datatype(self, value):
        self._datatype = value

    def collect_data(self, data):
        data.update({'name': str(self.name)})
        data.update({'value': self.value})
        data.update({'datatype': str(self.datatype)})
        return data

    def save(self):
        """
        Save variable data on filesystem.

        :ivar str setup_path: Setup path.
        :return: Dictionary containing the variable info to save.
        :rtype: dict
        """
        data2save = {}
        data2save.update({'name': str(self.name)})
        data2save.update({'value': self.value})
        data2save.update({'datatype': str(self.datatype)})
        return data2save

    def load(self, data):
        """
        Load variable data from filesystem

        :ivar dict data: data object that contains all task variable info
        """
        self.name = data['name']
        self.value = data['value']
        self.datatype = data['datatype']



import os
import uuid

from pybpodgui_api.utils.generate_name import generate_name


class UserBase():

    def __init__(self, _project):
        self._path = None
        self.uuid4 = uuid.uuid4()
        self.project = _project
        self._name = generate_name([x.name for x in _project.users], "user")
        self.name = self._name
        self.connection = 'local'

        self.project += self

    @property
    def name(self):
        """
        Get and set setup name

        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, value):
        self._name = value
    
    @property
    def path(self):
        if self.project.path is None:
            return None
        return os.path.join(self.project.path, 'users', self.name)

    def __unicode__(self):
        return self.name

    def __str__(self):
        return self.__unicode__()



# !/usr/bin/python3
# -*- coding: utf-8 -*-
from confapp import conf
from pybpodgui_plugin.models.project.project_uibusy import ProjectUIBusy

Project = type(
    'Project',
    tuple(conf.GENERIC_EDITOR_PLUGINS_FINDER.find_class('models.project.Project') + [ProjectUIBusy]),
    {}
)



from threading import Thread, Event
from queue import Queue, Empty
import socket


class NonBlockingSocketReceive:
    def __init__(self, sck):
        """
        stream: the stream to read from.
                Usually a process' stdout or stderr.
        """
        self._s = sck
        self._q = Queue()

        class PopulateQueue(Thread):
            def __init__(self, sck, queue):
                Thread.__init__(self)
                self.daemon = True
                self.socket = sck
                self.queue = queue
                self.event = Event()

            def run(self):
                try:
                    self.socket.settimeout(1.0)
                    data = None
                    while True:
                        if self.event.is_set():
                            break

                        if not self.socket:
                            break
                        try:
                            data = self.socket.recv(64)
                            if not data:
                                self.event.set()
                        except socket.timeout:
                            pass
                        if data:
                            self.queue.put(data)
                            data = None

                        self.event.wait(0.01)

                except OSError:
                    self.event.set()

        self._t = PopulateQueue(self._s, self._q)
        self._t.daemon = True
        self._t.start()  # start collecting lines from the stream

    def readline(self, timeout=None):
        try:
            return self._q.get(block=timeout is not None, timeout=timeout)
        except Empty:
            return None

    def close(self):
        self._t.event.set()

    def is_alive(self):
        return self._t.is_alive()


class UnexpectedEndOfStream(Exception):
    pass



# !/usr/bin/python3
# -*- coding: utf-8 -*-

import logging
from pybpodapi.com.messaging.base_message import BaseMessage
from pybpodapi.utils import date_parser

logger = logging.getLogger(__name__)


class StateOccurrence(BaseMessage):
    """
    Store timestamps for a specific state occurrence of the state machine

    :ivar str name: name of the state
    :ivar list(StateDuration) timestamps: a list of timestamps (start and end) that corresponds to occurrences of this state
    """

    MESSAGE_TYPE_ALIAS = "STATE"
    MESSAGE_COLOR = (0, 100, 0)

    def __init__(self, state_name, host_timestamp, end_timestamp):
        """

        :param str name: name of the state
        """
        super(StateOccurrence, self).__init__(state_name, host_timestamp)

        self.start_timestamp = host_timestamp
        self.end_timestamp = end_timestamp

    def tolist(self):
        return [
            self.MESSAGE_TYPE_ALIAS,
            str(self.pc_timestamp),
            self.start_timestamp,
            self.end_timestamp,
            self.content,
            self.end_timestamp - self.start_timestamp,
        ]

    @classmethod
    def fromlist(cls, row):
        """
        Returns True if the typestr represents the class
        """
        obj = cls(
            row[4], float(row[2]) if row[2] else None, float(row[3]) if row[3] else None
        )
        obj.pc_timestamp = date_parser.parse(row[1])

        return obj

    @property
    def state_name(self):
        return self.content



# !/usr/bin/python3
# -*- coding: utf-8 -*-
import loggingbootstrap

from confapp import conf

__version__ = "1.8.2"
__author__ = "Carlos Mao de Ferro"
__credits__ = ["Carlos Mao de Ferro", "Ricardo Ribeiro", "Sérgio Copeto", 'Luís Teixeira']
__license__ = "MIT"
__maintainer__ = ["Carlos Mao de Ferro", "Ricardo Ribeiro", "Sérgio Copeto", 'Luís Teixeira']
__email__ = ["cajomferro@gmail.com", "ricardojvr@gmail.com", "sergio.copeto@research.fchampalimaud.org", 'micboucinha@gmail.com']
__status__ = "Development"

conf += 'pybpodgui_plugin.settings'
conf += 'pybpodgui_plugin.resources'

# setup different loggers but output to single file
loggingbootstrap.create_double_logger("pybpodgui_plugin", conf.APP_LOG_HANDLER_CONSOLE_LEVEL,
                                      conf.APP_LOG_FILENAME,
                                      conf.APP_LOG_HANDLER_FILE_LEVEL)

loggingbootstrap.create_double_logger("pybranch", conf.APP_LOG_HANDLER_CONSOLE_LEVEL,
                                      conf.APP_LOG_FILENAME,
                                      conf.APP_LOG_HANDLER_FILE_LEVEL)

loggingbootstrap.create_double_logger("pybpodapi", conf.APP_LOG_HANDLER_CONSOLE_LEVEL,
                                      conf.APP_LOG_FILENAME,
                                      conf.APP_LOG_HANDLER_FILE_LEVEL)

if conf.USE_MULTIPROCESSING:
    # https://docs.python.org/3.5/library/multiprocessing.html#multiprocessing.freeze_support
    from multiprocessing import freeze_support  # @UnresolvedImport
    freeze_support()



from pybpod_gui_plugin_emulator import EmulatorGUI
from pybpodgui_plugin.models.setup.setup_dockwindow import SetupDockWindow
from pyforms_gui.controls.control_button import ControlButton
from confapp import conf


class SetupWindowEmulator(SetupDockWindow):
    """
    Setup

    :param experiment: The parent Experiment of this Setup
    """

    def __init__(self, experiment=None):
        super(SetupWindowEmulator, self).__init__(experiment)

        self._emulator = ControlButton('Test protocol IO',
                                       default=self.__emulator_btn_evt,
                                       icon=conf.EMULATOR_LAUNCH_ICON,
                                       enabled=False)

        self._formset = [
            '_name',
            '_board',
            '_task',
            '_emulator',
            '_detached',
            ('_run_task_btn', '_kill_task_btn'),
            ('_stoptrial_btn', '_pause_btn'),
            #'=',
            {
                'Subjects': [
                    # '_allsubjects',
                    '',
                    '_add_subject',
                    '_subjects_list',
                ],
                'Variables': [
                    '_varspanel',
                ],
            }
        ]

    def show(self):
        super().show()
        if self._board.value and self._task.value:
            self._emulator.enabled = True

    def _task_changed_evt(self):
        if hasattr(self, '_emulator'):
            self._emulator.enabled = (self._task.value != 0 and self._board.value != 0)
        if hasattr(self, 'emulator_plugin') and self._task.value:
            self.emulator_plugin.update_task(self._task.value)
        super()._task_changed_evt()

    def _board_changed_evt(self):
        if hasattr(self, '_emulator'):
            self._emulator.enabled = (self._board.value != 0 and self._task.value != 0)
        if hasattr(self, 'emulator_plugin') and self._board.value:
            self.emulator_plugin.update_board(self._board.value)
        super()._board_changed_evt()

    def __emulator_btn_evt(self):
        if hasattr(self, 'emulator_plugin') and self.emulator_plugin.started_correctly:
            self.emulator_plugin.show()
        else:
            self.create_emulator_window()

    def create_emulator_window(self):
        self.emulator_plugin = EmulatorGUI(self)
        if self.emulator_plugin.started_correctly:
            self.emulator_plugin.show()
            self.emulator_plugin.resize(*conf.EMULATOR_PLUGIN_WINDOW_SIZE)



# !/usr/bin/python3
# -*- coding: utf-8 -*-

from confapp import conf
from pybpodgui_plugin.models.experiment.experiment_uibusy import ExperimentUIBusy

Experiment = type(
    'Experiment',
    tuple(conf.GENERIC_EDITOR_PLUGINS_FINDER.find_class('models.experiment.Experiment') + [ExperimentUIBusy]),
    {}
)



from pybpodapi.settings import TARGET_BPOD_FIRMWARE_VERSION

if TARGET_BPOD_FIRMWARE_VERSION == "17":
    from pybpodapi.bpod.hardware.input_channels.bpod0_7_9_fw13 import InputName
elif TARGET_BPOD_FIRMWARE_VERSION == "15":
    from pybpodapi.bpod.hardware.input_channels.bpod0_7_9_fw13 import InputName
elif TARGET_BPOD_FIRMWARE_VERSION == "13":
    from pybpodapi.bpod.hardware.input_channels.bpod0_7_9_fw13 import InputName
elif TARGET_BPOD_FIRMWARE_VERSION == "9":
    from pybpodapi.bpod.hardware.input_channels.bpod0_7_5_fw9 import InputName
else:
    from pybpodapi.bpod.hardware.input_channels.bpod0_7_5_fw9 import InputName



from pyforms.basewidget import BaseWidget
from pyforms.controls   import ControlLabel
from pyforms.controls   import ControlText
from pyforms.controls   import ControlList
from pyforms.controls   import ControlButton
from pyforms.controls   import ControlProgress
from pyforms.controls   import ControlTextArea
from pyforms.controls   import ControlCombo
from pyforms.controls   import ControlWeb

from AnyQt.QtWidgets import QApplication
from AnyQt.QtCore import QTimer

from confapp import conf

from importlib import reload

import yaml
import sys
import subprocess
import xmlrpc.client
import pkg_resources
import datetime



class PackageWindow(BaseWidget):

    pypi = xmlrpc.client.ServerProxy('https://pypi.org')

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


        local = self.get_current_distribution(self.title)
        new_version, all_versions, description = self.get_pypi_distribution(self.title)
        
        current_version = local.version if local is not None else None

        nlines = len(description.split('\n'))

        self.set_margin(10)
        self.setMinimumHeight(160 if nlines<3 else nlines*20)

        self._version = ControlLabel(
            'Installed version {0}'.format(current_version) if new_version!=current_version else 'Latest installed', 
            visible=current_version is not None, 
            style='font-size:11pt;max-width:200px;min-height:30px;'+ ('color:blue;' if new_version==current_version else 'color:#FF7F50')
        )
        
        self._versions = ControlCombo(
            'Version', 
            items=[ [v] for v in all_versions],
            style='max-width:200px;min-height:30px;'
        )
        
        self._label = ControlLabel(
            self.title, 
            style='font-weight: bold;min-height:30px;font-size:13pt;vertical-align: middle;min-width:200px;'
        )
        
        self._install_btn  = ControlButton(
            'Update' if current_version else 'Install', 
            style='max-width:100px;min-height:30px;'+ ('color:auto;' if current_version is None or new_version==current_version else 'color:#FF7F50'),
            default=self.__install_btn_evt
        )

        self._description  = ControlTextArea(
            default=description, 
            readonly=True, 
            style='background-color: #333; color:white; margin-bottom: 20px;'
        )
        
        self.formset = [
            ('_label','_version',' ','_versions','_install_btn'),
            '_description'
        ]

        """
        if new_version is not False and current_version is False:
            self._install_btn.label = 'Install version {0}'.format(new_version)
        elif new_version is not False and current_version is not False:
            self._install_btn.label = 'Update from {0} to {1}'.format(current_version, new_version)
        else:
            self._install_btn.enabled = False
            self._install_btn.label = 'No updates available'
        """

    def get_current_distribution(self, name):
        try:
            mod = sys.modules.get(name, None)
            if mod: 
                reload(mod)
            dist = pkg_resources.get_distribution(name)
        except pkg_resources.DistributionNotFound:
            dist = None
        return dist



    def get_pypi_distribution(self, name):

        new_version = self.pypi.package_releases(name)
        if not new_version:
            new_version = self.pypi.package_releases(name.capitalize())

        if new_version is None: return new_version

        new_version  = new_version[0]
        all_versions = self.pypi.package_releases(name, True)
        data = self.pypi.release_data(name, new_version)


        return new_version, all_versions, data.get('summary', '')



    def __install_btn_evt(self):
        subprocess.check_call([sys.executable, '-m', 'pip', 'install', self.title+'=='+self._versions.value])
        self.parent_widget.load_packages_list()



from pybpodapi.bpod_modules.bpod_module import BpodModule


class WavePlayer(BpodModule):
    
    def __init__(self):
        print('INITING')

    @staticmethod
    def check_module_type(module_name):
        return module_name and module_name.startswith('WavePlayer')


# !/usr/bin/python3
# -*- coding: utf-8 -*-

import logging

logger = logging.getLogger(__name__)


class InputName(object):
    """
    Input channels available on Bpod box
    These values must be set according to Bpod firmware specification.
    """

    #: Serial1
    Serial1 = "Serial1"

    #: Serial2
    Serial2 = "Serial2"

    #: Serial1_3
    Serial3 = "Serial3"

    #: Serial1_4
    USB1 = "USB1"

    #: BNC1
    BNC1 = "BNC1"

    #: BNC2
    BNC2 = "BNC2"

    #: Wire1
    Wire1 = "Wire1"

    #: Wire2
    Wire2 = "Wire2"

    #: Input port 1
    Port1 = "Port1"

    #: Input port 2
    Port2 = "Port2"

    #: Input port 3
    Port3 = "Port3"

    #: Input port 4
    Port4 = "Port4"

    #: Input port 5
    Port5 = "Port5"

    #: Input port 6
    Port6 = "Port6"

    #: Input port 7
    Port7 = "Port7"

    #: Input port 8
    Port8 = "Port8"



from pybpod_alyx_module.module_gui import AlyxModuleGUI
from pybpodgui_plugin.models.project.project_window import ProjectWindow


class ProjectTreeNode(ProjectWindow):

	def create_treenode(self, tree):
		node = super(ProjectTreeNode, self).create_treenode(tree)
		self.open_alyx_action = tree.add_popup_menu_option(
            'Sync to Alyx',
            self.open_alyx_window,
            item= node
            )

		return node
	
	def open_alyx_window(self):
		if not hasattr(self, 'alyx_window'):
			self.alyx_window = AlyxModuleGUI(self)
		self.alyx_window.show()
		return self.alyx_window



# !/usr/bin/python3
# -*- coding: utf-8 -*-

import logging

from pybpodgui_plugin.models.task.windows.code_editor import CodeEditor
from pybpodgui_plugin.models.task.task_treenode import TaskTreeNode

logger = logging.getLogger(__name__)


class TaskDockWindow(TaskTreeNode):
    """
    Dock window settings.
    Define here behaviors associated with board dock window.

    **Properties**

        mainwindow
            Returns project main window.

    **Methods**

    """

    def show(self):
        """
        Select this window as the main window on the details section.
        """
        self.mainwindow.details.value = self
        super(TaskDockWindow, self).show()

    def focus_name(self):
        """
        Sets interface focus on the task name text field
        """
        self._namefield.form.lineEdit.setFocus()

    def remove(self):
        """

        Prompts user to confirm task removal and closes mdi windows associated with this task.

        .. seealso::
            This method extends task tree node :py:meth:`pybpodgui_plugin.models.task.task_treenode.TaskTreeNode.remove`.

        """
        reply = self.question('Task "{0}" will be deleted. Are you sure?'.format(self.name), 'Warning')
        if reply == 'yes':
            if hasattr(self, '_code_editor'):
                self.mainwindow.mdi_area -= self._code_editor
            super(TaskDockWindow, self).remove()

    def edit_btn_evt(self):
        """
        Open code editor window on the mdi section for the task source code.

        .. seealso::
            This event may be fired on:
                * Double click event (tree node): :py:meth:`pybpodgui_plugin.models.task.task_treenode.TaskTreeNode.node_double_clicked_event`.
                * Key press event (tree node): :py:meth:`pybpodgui_plugin.models.task.task_treenode.TaskTreeNode.node_key_pressed_event`.
        """
        if self.project.path is None or self.filepath is None:
            self.warning("The task was not commited yet,<br/>please save the project first to create the task file.",
                         "Cannot edit the file yet.")
        else:
            try:
                if not hasattr(self, '_code_editor'):
                    self._code_editor = CodeEditor(self)
                self.mainwindow.mdi_area += self._code_editor
            except FileNotFoundError as err:
                logger.warning(str(err))
                self.warning(
                    "Task file does not exist yet.\nPlease save the project to create the task file.",
                    "Cannot edit the file yet.")

    @property
    def mainwindow(self):
        return self.project.mainwindow



# !/usr/bin/python
# -*- coding: utf-8 -*-

import logging

from confapp import conf

from AnyQt.QtGui import QIcon
from pyforms.controls import ControlTree

from pybpodgui_plugin_trial_timeline.trial_timeline import TrialTimeline

logger = logging.getLogger(__name__)

class SessionTreeNode(object):

    def create_treenode(self, tree):
        node = super(SessionTreeNode, self).create_treenode(tree)

        self.trialtimeline_action = tree.add_popup_menu_option(
            'Trial Timeline', 
            self.open_trialtimeline_window,
            item=self.node
            )

        self.trialtimeline_detached_action = tree.add_popup_menu_option(
            'Trial Timeline (Detached)', 
            self.open_trialtimeline_window_detached,
            item=self.node
            )

        return node

    def open_trialtimeline_window(self):
        
        self.load_contents()

        #does not show the window if the detached window is visible
        if hasattr(self, 'trial_timeline_win_detached') and self.trial_timeline_win_detached.visible: return 

        if not hasattr(self,'trial_timeline_win'):
            self.trial_timeline_win = TrialTimeline(self)
            self.trial_timeline_win.show()
        else:
            self.trial_timeline_win.show()
    
    def open_trialtimeline_window_detached(self):
        
        self.load_contents()

        if hasattr(self, 'trial_timeline_win') and self.trial_timeline_win.visible: return 
     
        if not hasattr(self,'trial_timeline_win_detached'):
            self.trial_timeline_win_detached = TrialTimeline(self)
            self.trial_timeline_win_detached.show(True)
        else:
            self.trial_timeline_win_detached.show(True)
    
    @property
    def name(self):
        return super(SessionTreeNode, self.__class__).name.fget(self)

    @name.setter
    def name(self, value):
        super(SessionTreeNode, self.__class__).name.fset(self, value)
        if hasattr(self, 'trialsplot_win'):
            self.trialsplot_win.title = value 
    




from pybpodapi.com.arcom import ArCOM, ArduinoTypes
from pybpodapi.bpod_modules.bpod_module import BpodModule
import struct
import numpy as np
import wave
import time

class WavePlayerModule(object):
    
    COM_HANDSHAKE            = 227
    COM_PLAY_WAVEFORM        = ord('P')
    COM_STOP_PLAYBACK        = ord('X')
    COM_LOAD_WAVEFORM        = ord('L')
    COM_SET_OUTPUT_RANGE     = ord('R')
    COM_SET_SAMPLING_PERIOD  = ord('S')
    COM_SET_LOOP             = ord('O')
    COM_SET_BPOD_EVENTS      = ord('V')
    COM_SET_TRIGGER          = ord('T')
    COM_LOAD_TRIGGER_PROFILE = ord('F')
    COM_GET_PARAMETERS       = ord('N')
    
    CURRENT_FIRMWARE_VERSION = 1

    TRIGGER_MODE_NORMAL = 0
    TRIGGER_MODE_MASTER = 1
    TRIGGER_MODE_TOGGLE = 2

    RANGE_VOLTS = ['0V:5V', '0V:10V', '0V:12V', '-5V:5V', '-10V:10V', '-12V:12V']
    RANGE_VOLTS_0_5  = 0
    RANGE_VOLTS_0_10 = 1
    RANGE_VOLTS_0_12 = 2
    RANGE_VOLTS_MINUS5_5   = 3
    RANGE_VOLTS_MINUS10_10 = 4
    RANGE_VOLTS_MINUS12_12 = 5


    def __init__(self, serial_port = None):

        self.is_open = False

        if serial_port:
            self.open(serial_port)



    def open(self, serialport):

        self.arcom = ArCOM().open(serialport,115200)

        self.arcom.write_array( ArduinoTypes.get_uint8_array([self.COM_HANDSHAKE]) )

        ack = self.arcom.read_uint8()
        if ack!=228:
            raise Exception('Could not connect =( ')
        
        version = self.arcom.read_uint32()
        if version < self.CURRENT_FIRMWARE_VERSION:
            raise Exception("""Error: old firmware detected - v{0}.
                The current version is: {1}.
                Please update the I2C messenger firmware using Arduino.""".format(
                version, self.CURRENT_FIRMWARE_VERSION
            ))
        self.firmware_version = version
        
        self.get_parameters()
        self.is_open = True

        return True


    def disconnect(self):
        if hasattr(self,'arcom'):
            self.arcom.close()
            del self.arcom
        self.is_open = False


    def play(self, channels, wave_index):
        self.arcom.write_array(ArduinoTypes.get_uint8_array([self.COM_PLAY_WAVEFORM, channels, wave_index]))


    def stop(self):
        self.arcom.write_array(ArduinoTypes.get_uint8_array([self.COM_STOP_PLAYBACK]))


    def load_waveform(self, wave_index, wavform):
        
        data2send  = ArduinoTypes.get_uint8_array([self.COM_LOAD_WAVEFORM, wave_index])
        data2send += ArduinoTypes.get_uint32_array([len(wavform)])

        wavform = np.array(wavform)

        if self.output_range==0:
            positive_only = 1.0
            voltage_width = 5.0
        elif self.output_range==1:
            positive_only = 1.0
            voltage_width = 10.0
        elif self.output_range==2:
            positive_only = 1.0
            voltage_width = 12.0
        elif self.output_range==3:
            positive_only = 0.0
            voltage_width = 10.0
        elif self.output_range==4:
            positive_only = 0.0
            voltage_width = 20.0
        elif self.output_range==5:
            positive_only = 0.0
            voltage_width = 24.0 
        else:
            positive_only = 0.0
            voltage_width = 10.0
        
        min_wave  = min(wavform);
        max_wave  = max(wavform);
        max_range = voltage_width+(positive_only*0.5);
        min_range = ( (voltage_width/2.0) * -1.0 ) * ( 1.0 - positive_only )

        if (min_wave < min_range) or (max_wave > max_range):
            raise Exception("""Error setting waveform: All voltages must be within the current range: {0}.""".format(
                self.RANGE_VOLTS[self.output_range]
            ))
        
        offset = (voltage_width/2.0)*(1.0-positive_only)
        wavbits = np.ceil(((wavform+offset)/voltage_width)*( (2.0**16.0) -1.0));

        data2send += ArduinoTypes.get_uint16_array(wavbits)

        self.arcom.write_array(data2send)
        
        ack = self.arcom.read_uint8()
        return ack==1


    def set_output_range(self, value):
        self.arcom.write_array(ArduinoTypes.get_uint8_array( [self.COM_SET_OUTPUT_RANGE, value] ))
        ack = self.arcom.read_uint8()
        return ack==1

    def set_sampling_period(self, value):

        if value < 1 or value > 200000:
            raise Exception('Error setting sampling rate: valid rates are in range: [1;200000] Hz')
            
        sampling_period_microseconds = (1.0/value)*1000000.0

        data2send  = ArduinoTypes.get_uint8_array([self.COM_SET_SAMPLING_PERIOD])
        data2send += ArduinoTypes.get_float(sampling_period_microseconds)
        
        self.arcom.write_array(data2send)
       

    def set_loop_duration(self, loop_duration):

        if len(self.loop_mode)!=self.n_channels:
            raise Exception('Wrong loop modes')

        if len(loop_duration)!=self.n_channels:
            raise Exception('Wrong loop durations')

        bytes2send =  ArduinoTypes.get_uint8_array([self.COM_SET_LOOP] + self.loop_mode)
        bytes2send += ArduinoTypes.get_uint32_array(np.array(loop_duration))

        self.arcom.write_array(bytes2send)

        ack = self.arcom.read_uint8()
        if ack==1: self._loop_duration = loop_duration
        return ack==1



    def set_loop_mode(self, loop_mode):

        if len(loop_mode)!=self.n_channels:
            raise Exception('Wrong loop modes')

        if len(self.loop_duration)!=self.n_channels:
            raise Exception('Wrong loop durations')

        for i, mode in enumerate(loop_mode):
            if mode and self.loop_duration[i]==0:
                raise Exception('Error: before enabling loop mode, each enabled channel must have a valid loop duration.')

        bytes2send =  ArduinoTypes.get_uint8_array([self.COM_SET_LOOP] + loop_mode)
        bytes2send += ArduinoTypes.get_uint32_array(self.loop_duration)

        self.arcom.write_array(bytes2send)

        ack = self.arcom.read_uint8()
        if ack==1: self._loop_mode = loop_mode
        return ack==1


    def set_bpod_events(self, values):
        bytes2send = ArduinoTypes.get_uint8_array(
            [self.COM_SET_BPOD_EVENTS] + values
        )
        self.arcom.write_array(bytes2send)
        ack = self.arcom.read_uint8()

        if ack==1:
            self._bpod_events = values
        return ack==1
        
    def set_trigger_mode(self, value):
        self.arcom.write_array(
            ArduinoTypes.get_uint8_array([self.COM_SET_TRIGGER, value])
        )
        ack = self.arcom.read_uint8()
        
        if ack==1:
            self._trigger_mode = value
        return ack==1
        

    def set_trigger_profiles(self, trigger_profiles):

        length, width = trigger_profiles.shape

        if length!=self.n_trigger_profiles or width!=self.n_channels:
            raise Exception(
                'Error setting trigger profiles: matrix of trigger profiles must be {0} profiles X {1} channels.'.format(
                self.n_trigger_profiles, self.n_channels
            ))

        """ TODO
        if sum(sum((profileMatrix > 0)') > obj.maxSimultaneousChannels) > 0
                error(['Error setting trigger profiles: the current sampling rate only allows ' num2str(obj.maxSimultaneousChannels) ' channels to be triggered simultaneously. Your profile matrix contains at least 1 profile with too many channels.']);
            end
        """
        profile_matrix_out = trigger_profiles
        profile_matrix_out[profile_matrix_out==0] = 256
        
        self.arcom.write_array(
            ArduinoTypes.get_uint8_array([self.COM_LOAD_TRIGGER_PROFILE, profile_matrix_out-1])
        )
        ack = self.arcom.read_uint8()
        
        if ack==1:
            self._trigger_profiles = trigger_profiles
        return ack==1
         
            



    ######################################################################
    ### PROPERTIES #######################################################
    ######################################################################

    
    @property
    def trigger_mode(self):
        return self._trigger_mode
    
    @property
    def bpod_events(self):
        return self._bpod_events
    
    @property
    def loop_mode(self):
        return self._loop_mode
    
    @property
    def loop_duration(self):
        return self._loop_duration
        
    @property
    def trigger_profile_enable(self):
        return self._trigger_profile_enable
    
    @property
    def trigger_profiles(self):
        return self._trigger_profiles
    
    @property
    def output_range(self):
        return self._output_range
    
    @property
    def sampling_rate(self):
        return self._sampling_rate

    @property
    def max_waves(self):
        return self._max_waves
    
    @property
    def n_channels(self):
        return self._n_channels

    @property
    def n_trigger_profiles(self):
        return self._n_trigger_profiles
    

    ######################################################################
    ### PRIVATE FUNCTIONS ################################################
    ######################################################################

    def get_parameters(self):
        self.arcom.write_array([self.COM_GET_PARAMETERS])
        # number of output channels
        self._n_channels             = self.arcom.read_uint8()
        # maximum number of waveforms supported
        self._max_waves              = self.arcom.read_uint16()
        # current trigger mode
        self._trigger_mode           = self.arcom.read_uint8()
        # 0 = standard trigger mode, 1 = trigger profile mode
        self._trigger_profile_enable = self.arcom.read_uint8()==1
        # maximum number of trigger profiles supported
        self._n_trigger_profiles     = self.arcom.read_uint8()
        # index of the currently selected range 
        self._output_range           = self.arcom.read_uint8()
        # sampling period (in microseconds)
        sampling_period_microseconds = self.arcom.read_float32()
        # 0 = off, 1 = bpod event reporting
        self._bpod_events   = [v==1 for v in self.arcom.read_uint8_array(self.n_channels)]
        
        # 0 = off, 1 = on
        self._loop_mode     = [v==1 for v in self.arcom.read_uint8_array(self.n_channels)]

        self._sampling_rate = round((1/sampling_period_microseconds)*1000000)

        # duration of loop playback in samples
        self._loop_duration = np.array(self.arcom.read_uint32_array(self.n_channels))*self.sampling_rate
        
        #self.is_playing       = [False for i in range(self.n_channels)]
        self._trigger_profiles = np.zeros( (self.n_trigger_profiles, self.n_channels) )
        



    def _print_parameters(self):
        print('Number of channels:',self.n_channels)
        print('Max waves:',self.max_waves)
        print('Trigger mode:',self.trigger_mode)
        print('Trigger profile enabled:',self.trigger_profile_enable)
        print('Number of trigger profiles:',self.n_trigger_profiles)
        print('Output range:',self.output_range)
        print('Bpod events:',self.bpod_events)
        print('Loop mode',self.loop_mode)
        print('Sampling rates:',self.sampling_rate)
        print('Loop durations:',self.loop_duration)


    def debug(self):
        self.get_parameters()
        self._print_parameters()


# !/usr/bin/python3
# -*- coding: utf-8 -*-

from pybpodgui_api.models.setup.task_variable import TaskVariable
from pyforms.controls import ControlCombo


class TaskVariableWindow(TaskVariable):

    def __init__(self, board_task, name=None, value=None, datatype='string'):
        self._varslist = board_task._vars

        self._combo = ControlCombo(None)
        self._combo.add_item('Number', 'number')
        self._combo.add_item('String', 'string')
        self._combo.value = datatype
        self._combo.changed_event = self.__datatype_changed_evt

        self._row_index = self._varslist.rows_count
        self._varslist += [name, self._combo, value]

        TaskVariable.__init__(self, board_task, name, value, datatype)

    def __datatype_changed_evt(self):
        datatype = self._combo.value
        if datatype == 'number':
            item = self._varslist.get_value(2, self._row_index)
            if not item.isnumeric():
                self._varslist.set_value(2, self._row_index, '0')

    def tolist(self):
        return [self.name, self._combo, self.value]

    @property
    def name(self):
        return self._varslist.get_value(0, self._row_index)

    @name.setter
    def name(self, value):
        self._varslist.set_value(0, self._row_index, value)

    @property
    def value(self):
        value = self._varslist.get_value(2, self._row_index)
        return float(value) if self.datatype == 'number' else value

    @value.setter
    def value(self, value): self._varslist.set_value(2, self._row_index, value)

    @property
    def datatype(self): return self._varslist.get_value(1, self._row_index).value

    @datatype.setter
    def datatype(self, value):  self._varslist.get_value(1, self._row_index).value = value



import logging
from confapp import conf
from datetime import datetime as datetime_now

import pybpodapi
from pybpodapi.com.messaging.trial import Trial
from pybpodapi.com.messaging.state_occurrence import StateOccurrence

from sca.formats import csv

import sys
from pybpodapi.com.stdout_buffer import StdoutBuffer
from pybpodapi.com.stderr_buffer import StderrBuffer

logger = logging.getLogger(__name__)


class StreamsWrapper(object):
    def __init__(self, streams):
        self.streams = streams

    def write(self, data):
        for stream in self.streams:
            stream.write(data)

    def flush(self):
        for stream in self.streams:
            stream.flush()

    def close(self):
        for stream in self.streams:
            stream.flush()
            stream.close()


class Session(object):
    """
    Stores information about bpod run, including the list of trials.

    :ivar list(Trial) trials: a list of trials
    :ivar int firmware_version: firmware version of Bpod when experiment was run
    :ivar int bpod_version: version of Bpod hardware when experiment was run
    :ivar datetime start_timestamp: it stores session start timestamp

    """

    MSGTYPE_DEBUG = "debug"
    MSGTYPE_ENDTRIAL = "END-TRIAL"
    MSGTYPE_ERROR = "error"
    MSGTYPE_INFO = "INFO"
    MSGTYPE_SOFTCODE = "SOFTCODE"
    MSGTYPE_STDERR = "stderr"
    MSGTYPE_STDOUT = "stdout"
    MSGTYPE_TRIAL = "TRIAL"
    MSGTYPE_WARNING = "warning"
    MSGTYPE_STATE = "STATE"
    MSGTYPE_TRANSITION = "TRANSITION"

    INFO_SESSION_NAME = "SESSION-NAME"
    INFO_SESSION_STARTED = "SESSION-STARTED"
    INFO_SESSION_ENDED = "SESSION-ENDED"
    INFO_SERIAL_PORT = "SERIAL-PORT"
    INFO_NET_PORT = "NET-PORT"
    INFO_BPODAPI_VERSION = "BPOD-API-VERSION"

    INFO_PROTOCOL_NAME = "PROTOCOL-NAME"
    INFO_CREATOR_NAME = "CREATOR-NAME"
    INFO_PROJECT_NAME = "PROJECT-NAME"
    INFO_EXPERIMENT_NAME = "EXPERIMENT-NAME"
    INFO_BOARD_NAME = "BOARD-NAME"
    INFO_SETUP_NAME = "SETUP-NAME"
    INFO_SUBJECT_NAME = "SUBJECT-NAME"
    INFO_BPODGUI_VERSION = "BPOD-GUI-VERSION"

    INFO_TRIAL_BPODTIME = "TRIAL-BPOD-TIME"

    def __init__(self, path=None):
        self.ostdout = sys.stdout
        self.ostderr = sys.stderr

        # the variable will contain a list of streams where the session output
        # should be written.
        streams = []

        self.history = []  # type: list[Trial]
        self.trials = []  # type: list[Trial]
        self.firmware_version = None  # type: int
        self.bpod_version = None  # type: int
        self.start_timestamp = datetime_now.now()  # type: datetime

        self.csvwriter = None
        self._path = path

        # stream data to a file.
        if path:
            streams += [open(path, "w")]

        self.csvstream = StreamsWrapper(streams)
        self.csvwriter = csv.writer(
            self.csvstream,
            columns_headers=[
                "TYPE",
                "PC-TIME",
                "BPOD-INITIAL-TIME",
                "BPOD-FINAL-TIME",
                "MSG",
                "+INFO",
            ],
            software="PyBpod API v" + str(pybpodapi.__version__),
            def_url="http://pybpod-api.readthedocs.org",
            def_text="This file contains data recorded during a session from the PyBpod system",
        )

    def __del__(self):

        self.csvstream.close()

        sys.stdout = self.ostdout
        sys.stderr = self.ostderr

    def __add__(self, msg):
        """
        Add new trial to this session and associate a state machine to it

        :param pybpodapi.model.state_machine sma: state machine associated with this trial
        """

        if isinstance(msg, Trial):
            self.trials.append(msg)
        elif self.current_trial is not None:
            self.current_trial += msg

        self.history.append(msg)

        if self.csvwriter:
            self.csvwriter.writerow(msg.tolist())
            self.csvwriter.flush()

        return self

    def add_trial_events(self):

        current_trial = self.current_trial  # type: Trial
        sma = current_trial.sma

        visitedStates = [0] * current_trial.sma.total_states_added
        # determine unique states while preserving visited order
        uniqueStates = []
        nUniqueStates = 0
        uniqueStateIndexes = [0] * len(current_trial.states)

        for i in range(len(current_trial.states)):
            if current_trial.states[i] in uniqueStates:
                uniqueStateIndexes[i] = uniqueStates.index(current_trial.states[i])
            else:
                uniqueStateIndexes[i] = nUniqueStates
                nUniqueStates += 1
                uniqueStates.append(current_trial.states[i])
                visitedStates[current_trial.states[i]] = 1

        # Create a 2-d matrix for each state in a list
        uniqueStateDataMatrices = [[] for i in range(len(current_trial.states))]

        # Append one matrix for each unique state
        for i in range(len(current_trial.states)):
            if len(current_trial.state_timestamps) > 1:
                uniqueStateDataMatrices[uniqueStateIndexes[i]] += [
                    (
                        current_trial.state_timestamps[i],
                        current_trial.state_timestamps[i + 1],
                    )
                ]

        for i in range(nUniqueStates):
            thisStateName = sma.state_names[uniqueStates[i]]

            for state_dur in uniqueStateDataMatrices[i]:
                self += StateOccurrence(thisStateName, state_dur[0], state_dur[1])

        logger.debug("State names: %s", sma.state_names)
        logger.debug("nPossibleStates: %s", sma.total_states_added)
        for i in range(sma.total_states_added):
            thisStateName = sma.state_names[i]
            if not visitedStates[i]:
                self += StateOccurrence(thisStateName, float("NaN"), float("NaN"))

        logger.debug(
            "Trial states: %s",
            [str(state) for state in current_trial.states_occurrences],
        )

        # save events occurrences on trial
        # current_trial.events_occurrences = sma.raw_data.events_occurrences  # type: list

        logger.debug(
            "Trial events: %s",
            [str(event) for event in current_trial.events_occurrences],
        )

        logger.debug("Trial info: %s", str(current_trial))

    @property
    def current_trial(self):
        """
        Get current trial

        :rtype: Trial
        """
        return self.trials[-1] if len(self.trials) > 0 else None

    @current_trial.setter
    def current_trial(self, value):
        """
        Get current trial

        :rtype: Trial
        """
        self.trials[-1] = value



# !/usr/bin/python3
# -*- coding: utf-8 -*-

import logging
import re
from pybpodgui_api.models.task.task_io import TaskIO

logger = logging.getLogger(__name__)


class TaskFile(TaskIO):
    """ Represents a state machine """

    def find_task_variables_from_file(self):

        task_variables = []

        # match any line begining with v. and get variable name and value until end of line or # sign
        pattern = "^v\.(?P<vname>\w+)\s*\=\s*(?P<vvalue>.*(?=#)|.*)"

        with open(self.path, "r") as file:
            file_content = file.read()

        for v_name in re.findall(pattern, file_content, re.MULTILINE):
            if not v_name in [var_name for var_name in task_variables]:
                logger.debug("Find variable name: %s", v_name)
                task_variables.append(v_name)  # ignore variables value at this point

        return task_variables

    def find_states_from_file(self, start=1):
        return self.find_pattern(start, search_match="states")

    def find_events_from_file(self, start=1):
        return self.find_pattern(start, search_match="events")

    def find_pattern(self, start=1, search_match=None):

        matches = {}

        regex = r"{0}\s*=\s*\[(.+?)(?=\])".format(search_match)
        try:
            found_matches = re.findall(regex, self.code, flags=re.DOTALL)[0]
        except:
            logger.warning("Invalid {0} format or no {0} defined".format(search_match))
            return matches

        for idx, read_state in enumerate(found_matches.split(','), start=start):
            try:
                state_id = str(read_state.replace("'", ""))  # if it is string
            except:
                state_id = str(read_state.replace("'", ""))  # if it is a var

            matches[idx] = state_id.strip()

        logger.debug("Found %s: %s", search_match, matches)

        return matches



# !/usr/bin/python3
# -*- coding: utf-8 -*-

import logging

logger = logging.getLogger(__name__)


class Conditions(object):
    def __init__(self, max_states, n_conditions):
        self.matrix = [[] for i in range(max_states)]
        self.values = [0] * n_conditions
        self.channels = [0] * n_conditions
        self.max_size = n_conditions

    def get_max_index_used(self):

        for i in range(self.max_size - 1, -1, -1):
            if self.channels[i] != 0:
                return i

        return None



#!/usr/bin/python3
# -*- coding: utf-8 -*-

import os

SETTINGS_PRIORITY = 80


TIMELINE_PLUGIN_ICON = os.path.join(
    os.path.dirname(__file__), 'resources', 'graph.png')

TIMELINE_PLUGIN_WINDOW_SIZE = 650, 200
TIMELINE_PLUGIN_REFRESH_RATE = 100



from setuptools import setup, find_packages

PYTHON_VERSION_REQ = ">3.8.0"
IBLPYBPOD_CURRENT_VERSION = "2.0.2"

long_description = """
iblpybpod is a GUI application that enables interaction with the Bpod device from [Sanworks](https://sanworks.io/). Much credit 
and thanks go to the original creators of the [pybpod project](https://github.com/pybpod/pybpod).
"""

with open("requirements.txt") as f:
    require = [x.strip() for x in f.readlines() if not x.startswith('git+')]

setup(
    name="iblpybpod",
    version=IBLPYBPOD_CURRENT_VERSION,
    python_requires=PYTHON_VERSION_REQ,
    description="IBL implementation of pybpod software",
    license="MIT",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="IBL Staff",
    url="https://github.com/int-brain-lab/iblpybpod/",
    package_dir={"": "src"},
    packages=find_packages(where="src", exclude=['scratch', 'tests']),  # same as name
    include_package_data=True,
    install_requires=require,
    entry_points={"console_scripts": ["start-pybpod=pybpodgui_plugin.__main__:start"]},
    scripts=[]
)



# !/usr/bin/python3
# -*- coding: utf-8 -*-

import traceback
import logging
import os
import subprocess
import io
import sys
import json
import pandas as pd

from confapp import conf
from pathlib import Path

from pybpodapi.session import Session
from pybpodapi.utils import date_parser
from pybpodgui_api.models.board.board_io import BoardIO

from sca.formats.csv import CSV_DELIMITER, CSV_QUOTECHAR, CSV_QUOTING
import csv

from .non_blockingcsvreader import NonBlockingCSVReader
from .non_blockingstreamreader import NonBlockingStreamReader

logger = logging.getLogger(__name__)


class BoardCom(BoardIO):
    #### SETUP STATUS ####
    STATUS_READY = 0
    STATUS_RUNNING_TASK = 1  # The board is busy running a task

    INFO_CREATOR_NAME = 'CREATOR-NAME'
    INFO_PROJECT_NAME = 'PROJECT-NAME'
    INFO_EXPERIMENT_NAME = 'EXPERIMENT-NAME'
    INFO_BOARD_NAME = 'BOARD-NAME'
    INFO_SETUP_NAME = 'SETUP-NAME'
    INFO_SUBJECT_NAME = 'SUBJECT-NAME'
    INFO_BPODGUI_VERSION = 'BPOD-GUI-VERSION'

    def __init__(self, project):
        BoardIO.__init__(self, project)
        self.status = BoardCom.STATUS_READY

    ##########################################################################
    ####### PROPERTIES #######################################################
    ##########################################################################

    @property
    def status(self):
        """
        Get and set the board status

        :rtype: str
        """
        return self._status

    @status.setter
    def status(self, value):
        self._status = value

        # in case a session is running update the status of this sessions
        if not hasattr(self, '_running_session'):
            return

        session = self._running_session
        if value == self.STATUS_READY:
            session.status = session.STATUS_READY
        elif value == self.STATUS_RUNNING_TASK:
            session.status = session.STATUS_SESSION_RUNNING

    ##########################################################################
    ####### FUNCTIONS ########################################################
    ##########################################################################

    def pause_trial(self):
        if hasattr(self, 'proc'):
            self.proc.stdin.write("pause-trial\r\n".encode())
            self.proc.stdin.flush()

    def resume_trial(self):
        if hasattr(self, 'proc'):
            self.proc.stdin.write("resume-trial\r\n".encode())
            self.proc.stdin.flush()

    def stop_trial(self):
        if hasattr(self, 'proc'):
            self.proc.stdin.write("stop-trial\r\n".encode())
            self.proc.stdin.flush()

    def stop_task(self):
        if hasattr(self, 'proc'):
            self.proc.stdin.write("close\r\n".encode())
            self.proc.stdin.flush()
    
    def kill_task(self):
        if hasattr(self, 'proc'):
            self.proc.stdin.write("kill\r\n".encode())
            self.proc.stdin.flush()

    def log2board(self, data):
        """
        Function used to update the board log
        """
        pass

    def freegui(self):
        """
        Function used to release the processing to update the GUI events
        """
        pass

    def run_task(self, session, board_task, workspace_path, detached=False):
        """
        Run a task.

        :ivar Session session: Session to record the data.
        :ivar BoardTask board_task: Configuration to run session.
        :ivar str workspace_path: Not used. To be removed in the future.
        """
        session.subjects = [str([s.name, str(s.uuid4)]) for s in board_task.setup.subjects]

        xt_user = ''
        xt_subject = ''
        if hasattr(conf, 'PYBPOD_EXTRA_INFO'):
            if 'Users' in conf.PYBPOD_EXTRA_INFO:
                xt_user = session.user.toJSON()
            if 'Subjects' in conf.PYBPOD_EXTRA_INFO:
                xt_subject = []
                for sbj in board_task.setup.subjects:
                    xt_subject.append(sbj.toJSON())

        self._running_detached = detached
        self._running_boardtask = board_task
        self._running_task = board_task.task
        self._running_session = session

        board = board_task.board

        # create the session path
        Path(session.path).mkdir(parents=True, exist_ok=True)

        # load bpod configuration template
        template = os.path.join(os.path.dirname(__file__), 'run_settings_template.py')
        bpod_settings = open(template, 'r').read().format(
            serialport=board.serial_port,
            bnp_ports=('BPOD_BNC_PORTS_ENABLED = {0}'.format(board.enabled_bncports) if board.enabled_bncports else ''),
            wired_ports=('BPOD_WIRED_PORTS_ENABLED = {0}'.format(board.enabled_wiredports) if board.enabled_wiredports else ''),
            behavior_ports=('BPOD_BEHAVIOR_PORTS_ENABLED = {0}'.format(board.enabled_behaviorports) if board.enabled_behaviorports else ''),
            session_name=session.name,
            netport=board_task.board.net_port,
            stream2stdout=not detached,
            project=session.project.name,
            experiment=session.setup.experiment.name,
            board=board_task.board.name,
            setup=session.setup.name,
            session=session.name,
            protocolname=board_task.task.name,
            session_path=os.path.abspath(session.path).encode('unicode_escape').decode(),
            subjects=','.join(list(map(lambda x: '"'+str(x)+'"', session.subjects))),
            user=json.dumps([session.user.name, str(session.user.uuid4), session.user.connection] if session.user.uuid4 else None),
            subject_extra=','.join(list(map(lambda x: '"'+str(x)+'"', xt_subject))),
            user_extra=xt_user,
            variables_names=','.join(["'"+var.name+"'" for var in board_task.variables]),
            bpod_firmware_version=conf.TARGET_BPOD_FIRMWARE_VERSION
        )

        for var in board_task.variables:
            bpod_settings += "\n"+str(var)

        # create the bpod configuration file in the session folder
        settings_path = os.path.join(self._running_session.path, 'user_settings.py')
        with open(settings_path, 'w') as out:
            out.write(bpod_settings)

        # create the bpod configuration file in the session folder
        init_path = os.path.join(self._running_session.path, '__init__.py')
        with open(init_path, 'w') as out:
            pass

        ## Execute the PRE commands ###################################
        for cmd in board_task.task.commands:
            if cmd.when == 0:
                try:
                    cmd.execute(session=session)
                except Exception as err:
                    traceback.print_exc()
                    self.alert(str(err), "Unexpected error when executing a pre-command.")
        ###############################################################

        task = board_task.task

        self.start_run_task_handler()

        enviroment = os.environ.copy()
        enviroment['PYTHONPATH'] = os.pathsep.join([os.path.abspath(self._running_session.path)]+sys.path)

        if detached:
            self.proc = subprocess.Popen(
                ['python', os.path.abspath(task.filepath)],
                stdin=subprocess.PIPE,
                cwd=self._running_session.path,
                env=enviroment
            )
        else:
            session.data = pd.DataFrame(columns=['TYPE', 'PC-TIME', 'BPOD-INITIAL-TIME', 'BPOD-FINAL-TIME', 'MSG', '+INFO'])

            self.proc = subprocess.Popen(
                ['python', os.path.abspath(task.filepath)],
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                cwd=self._running_session.path,
                env=enviroment
            )
            self.csvreader = NonBlockingCSVReader(
                csv.reader(io.TextIOWrapper(self.proc.stdout, encoding='utf-8'),
                           delimiter=CSV_DELIMITER,
                           quotechar=CSV_QUOTECHAR,
                           quoting=CSV_QUOTING)
            )
            self.stderrstream = NonBlockingStreamReader(self.proc.stderr)

    def run_task_handler(self, flag=True):

        if not self._running_detached:
            row = self.csvreader.readline()
            data = ''
            while row is not None:
                if row is None:
                    break
                if len(row) == 6:
                    self._running_session.data.loc[len(self._running_session.data)] = row

                if self._running_session.uuid4 is None and len(row) == 2 and row[0] == '__UUID4__':
                    self._running_session.uuid4 = row[1]

                if len(row) > 0:
                    data += str(row)+'\n'

                self.freegui()
                row = self.csvreader.readline()

            if len(data) > 0:
                self.log2board(data)

            errline = self.stderrstream.readline()
            if errline is not None:
                self.log2board(errline)

        if flag and self.proc.poll() is not None:
            self.end_run_task_handler()

    def start_run_task_handler(self):
        self.status = self.STATUS_RUNNING_TASK

    def stop_thread(self):
        pass

    def end_run_task_handler(self):
        self.stop_thread()

        if not self._running_detached:
            # in case it is running detached
            errline = self.stderrstream.readline()
            if errline is not None:
                self.log2board(errline)

        # del self.proc

        session = self._running_session

        filepath = os.path.join(session.path, session.name+'.csv')
        session.filepath = filepath if os.path.exists(filepath) else None

        ## Execute the POST commands ##################################
        for cmd in self._running_task.commands:
            if cmd.when == 1:
                try:
                    cmd.execute(session=session)
                except Exception as err:
                    traceback.print_exc()
                    self.alert(str(err), "Unexpected error when executing the post-command")

        ###############################################################

        self.status = self.STATUS_READY

        if self._running_detached:
            session.load_contents()

        if session.data is not None:
            res = session.data.query("MSG=='{0}'".format(Session.INFO_SESSION_ENDED))
            for index, row in res.iterrows():
                session.ended = date_parser.parse(row['+INFO'])

            board_task = self._running_boardtask

            if board_task.update_variables:
                for var in board_task.variables:
                    res = session.data.query("TYPE=='VAL' and MSG=='{0}'".format(var.name))
                    for index, row in res.tail(1).iterrows():
                        value = row['+INFO']
                        if var.datatype == 'string':
                            var.value = value
                        else:
                            value.isdigit()
                            var.datatype == float(value)



#!/usr/bin/python3
# -*- coding: utf-8 -*-

import os

from confapp import conf


import pyforms
from pyforms.basewidget import BaseWidget

import logging

logger = logging.getLogger(__name__)

try:
	from pyforms.controls import ControlCodeEditor
except:
	logger.error("Could not import ControlCodeEditor. Is QScintilla installed?")


class GenericFileEditor(BaseWidget):
	def __init__(self, path, title, read_only=True):
		self._editor_title = title
		self._read_only = read_only

		BaseWidget.__init__(self, title)

		self.layout().setContentsMargins(5, 5, 5, 5)

		self.path = path

		self.editor = ControlCodeEditor(readonly=read_only)
		self.editor.value = self.content

		self.editor.changed_event = self.__content_changed_evt

	def set_cursor_position(self, line, index):
		self.editor._code_editor.setCursorPosition(line, index)

	def __content_changed_evt(self):
		"""
		
		:return: 
		"""
		self.content = self.editor.value

		self.info("You must restart project to apply changes.", "Restart required.")

		return True

	def beforeClose(self):
		""" 
		Before closing window, ask user if she wants to save (if there are changes)
		
		.. seealso::
			:py:meth:`pyforms.gui.Controls.ControlMdiArea.ControlMdiArea._subWindowClosed`.
		
		"""
		if self.editor.is_modified:
			reply = self.question('Save the changes', 'Save the file', buttons=['cancel','no', 'yes'])

			if reply=='yes':
				self.__content_changed_evt()

	@property
	def content(self):
		if not self.path or not os.path.exists(self.path):
			raise FileNotFoundError("User settings file not found!")
		with open(self.path, "r") as file:
			return file.read()
		return None

	@content.setter
	def content(self, value):
		with open(self.path, "w") as file:
			file.write(value)

	@property
	def title(self):
		return BaseWidget.title.fget(self)

	@title.setter
	def title(self, value):
		BaseWidget.title.fset(self, self._editor_title)


# Execute the application
if __name__ == "__main__":
	pyforms.start_app(SettingsEditor)



# !/usr/bin/python3
# -*- coding: utf-8 -*-

import sys
import os
import logging
from confapp import conf

import pkgutil

from AnyQt.QtWidgets import qApp, QApplication
from AnyQt.QtGui import QIcon

from pyforms.basewidget import BaseWidget
from pyforms.controls import ControlMdiArea
from pyforms.controls import ControlDockWidget

import pyforms_generic_editor
from pyforms_generic_editor.plugins import loader as plugin_loader
from pyforms_generic_editor.editor.generic_file_editor import GenericFileEditor

logger = logging.getLogger(__name__)


class BaseEditor(BaseWidget):
	def __init__(self):

		global conf
		conf += 'pyforms_generic_editor.resources'

		BaseWidget.__init__(self, conf.GENERIC_EDITOR_TITLE)

		self.mdi_area = ControlMdiArea()
		self.dock = ControlDockWidget(label='Projects', side=ControlDockWidget.SIDE_LEFT, order=1)
		self.details = ControlDockWidget(label='Details', side=ControlDockWidget.SIDE_RIGHT, order=2)

		self.dependencies = []
		self.plugins_classes = []

		self.mainmenu = [
			{'File': [
				{'Exit': self._exit_app, 'icon': QIcon(conf.EXIT_SMALL_ICON)}
			]},
			{'Window': [
				{'Cascade all': self._option_cascade_all},
				{'Tile all': self._option_tile_all},
				'-',
				{'Show projects': self.dock.show},
				{'Show details': self.details.show},
			]},
			{'Options': [
				{'Edit user settings': self._edit_user_settings, 'icon': QIcon(conf.USER_SETTINGS_ICON)}
			]},
			{'Help': [  # this option works for OSX also
				{'About QT': self._option_about_qt},
				{'About this project': self._option_about}
			]}
		]

		plugin_loader.install_plugins(self)  # Load plugins

		if conf.GENERIC_EDITOR_MODEL:
			if isinstance(conf.GENERIC_EDITOR_MODEL, str):
				c = self.__load_module(conf.GENERIC_EDITOR_MODEL)
			else:
				c = conf.GENERIC_EDITOR_MODEL
			self._model = c(self)

	def init_form(self):
		BaseWidget.init_form(self)

		# Loads a project automatically.
		if conf.DEFAULT_PROJECT_PATH:
			try:
				self.model.open_project(conf.DEFAULT_PROJECT_PATH)
			except Exception as err:
				for project in self.model.projects:
					project.close(silent=True)
				logger.warning(str(err), exc_info=True)
				self.warning("Please edit user settings.\n{0}".format(str(err)), "Invalid project path")

	def _option_close_all(self):
		self.mdi_area.value = []

	def _option_cascade_all(self):
		self.mdi_area.cascadeSubWindows()

	def _option_tile_all(self):
		self.mdi_area.tileSubWindows()

	def _option_about_qt(self):
		self.aboutQt("About Qt")

	def _option_about(self):
		plugins_info = ""
		for module in self.plugins_classes:
			try:
				version = module.__version__ if hasattr(module, '__version__') else 'Undefined'
				plugins_info += "<p>{0}: {1}</p>".format(str(module.__name__), version)
			except Exception as err:
				logger.error(str(err), exec_info=True)

		self.setWindowIcon(QIcon(conf.APP_LOGO_ICON))
		text = """
				<html><body>
				<h2>About {gui_title}</h2>
				<p>Pyforms Generic Editor: {version}</p>
				<h3>Scientific Software Platform (Champalimaud Foundation)</h3>
				<p>The Scientific Software Platform (SWP) from the Champalimaud Foundation provides technical know-how in software engineering and high quality software support for the Neuroscience and Cancer research community at the Champalimaud Foundation. We typical work on computer vision / tracking, behavioral experiments, image registration and database management.
				<p><a href="mailto:admin.it@neuro.fchampalimaud.org">admin.it@neuro.fchampalimaud.org</a></p>
				<h3>License</h3>
				<p>This is Open Source software. We use the <a href="https://www.gnu.org/licenses/gpl.html">GNU General Public License version 3</a>.</p>
				<h3>Installed plugins</h3>
				{plugins_info}
				</body></html>
				""".format(gui_title=self.title, version=pyforms_generic_editor.__version__, plugins_info=plugins_info)
		self.about(text, "About")

	def _edit_user_settings(self):
		"""
		Open code editor window on the mdi section for the task source code.

		.. seealso::
			This event may be fired on:
				* Double click event (tree node): :py:meth:`pybpodgui_plugin.models.task.task_treenode.TaskTreeNode.node_double_clicked_event`.
				* Key press event (tree node): :py:meth:`pybpodgui_plugin.models.task.task_treenode.TaskTreeNode.node_key_pressed_event`.
		"""

		try:
			if not hasattr(self, '_user_settings_editor'):  # this only happens the first time this window is opened
				user_settings_path = pkgutil.get_loader("user_settings").get_filename()
				logger.debug("User settings path: %s", user_settings_path)
				self._user_settings_editor = GenericFileEditor(user_settings_path, "User settings editor", False)

			self.mdi_area += self._user_settings_editor
		except Exception as err:
			logger.error(err, exc_info=True)
			self.warning(str(err), "Unexpected error")

	def __load_module(self, module_name):
		modules = module_name.split('.')
		moduleclass = __import__('.'.join(modules[:-1]), fromlist=[modules[-1]])
		return getattr(moduleclass, modules[-1])

	def closeEvent(self, event):
		""" 
		Confirms exit on close
		"""

		event.accept() if self._confirm_exit() else event.ignore()

	def _exit_app(self):
		"""
		Exit app option from the menu
		"""
		if self._confirm_exit():
			qApp.quit()

	def _confirm_exit(self):
		"""
		Prompts user to save projects changes and confirm exit
		
		:return: True to confirm exit and false otherwise
		:rtype: bool
		"""
		if self.model.is_saved():
			return True

		reply = self.question('Save all projects?', 'Save changes', buttons=['cancel','no','yes'])

		if reply=='yes':
			self.model.save_all_projects()
			return True
		elif reply=='no':
			return True
		else:
			return False

	@property
	def model(self):
		return self._model if hasattr(self, '_model') else None

	@model.setter
	def model(self, value):
		self._model = value



# !/usr/bin/python
# -*- coding: utf-8 -*-

import loggingbootstrap
import logging
import traceback
import sys
from importlib.util import find_spec

try:
	from confapp import conf

	# Initiating logging for pyforms. It has to be initiated manually here because we don't know yet
	# the logger filename as specified on settings
	loggingbootstrap.create_double_logger("pyforms", logging.INFO, 'app.log', logging.INFO)

except ImportError as err:
	logging.getLogger().critical(str(err), exc_info=True)
	exit("Could not load pyforms! Is it installed?")

# CHECK IF PYFORMS IS AVAILABLE WITHOUT ACTUALLY IMPORTING IT
pyforms_spec = find_spec("pyforms")
if not pyforms_spec:
	exit("Could not load pyforms! Is it installed?")

try:
	# pyforms is imported here first time through pyforms
	from pyforms_generic_editor import settings

	import user_settings

	conf += user_settings

	loggingbootstrap.create_double_logger("pyforms_generic_editor", conf.APP_LOG_HANDLER_CONSOLE_LEVEL,
	                                      conf.APP_LOG_FILENAME,
	                                      conf.APP_LOG_HANDLER_FILE_LEVEL)

	loggingbootstrap.create_double_logger("pyforms", conf.PYFORMS_LOG_HANDLER_CONSOLE_LEVEL, conf.APP_LOG_FILENAME,
	                                      conf.PYFORMS_LOG_HANDLER_FILE_LEVEL)

	# pyforms.controls is imported here first time
	from pyforms_generic_editor.editor.base_editor import BaseEditor as Editor

except Exception as err:
	exc_type, exc_value, exc_traceback = sys.exc_info()
	logging.getLogger("pyforms_generic_editor").critical(str(err), exc_info=True)
	conf.GENERIC_EDITOR_LOAD_EXCEPTION_TRACEBACK = traceback.format_exc()
	conf.GENERIC_EDITOR_LOAD_EXCEPTION_LINE = exc_traceback.tb_lineno
	settings.GENERIC_EDITOR_TITLE = 'PLEASE EDIT USER SETTINGS AND RESTART APP'
	from pyforms_generic_editor.editor.safe_mode_editor import SafeModeEditor as Editor


def start():
	import pyforms
	pyforms.start_app(Editor, conf.GENERIC_EDITOR_WINDOW_GEOMETRY)


if __name__ == '__main__':
	start()



from pybpodapi.settings import TARGET_BPOD_FIRMWARE_VERSION


if TARGET_BPOD_FIRMWARE_VERSION == "20":
    from pybpodapi.bpod.hardware.events.bpod0_7_9_fw20 import EventName
elif TARGET_BPOD_FIRMWARE_VERSION == "17":
    from pybpodapi.bpod.hardware.events.bpod0_7_9_fw13 import EventName
elif TARGET_BPOD_FIRMWARE_VERSION == "15":
    from pybpodapi.bpod.hardware.events.bpod0_7_9_fw13 import EventName
elif TARGET_BPOD_FIRMWARE_VERSION == "13":
    from pybpodapi.bpod.hardware.events.bpod0_7_9_fw13 import EventName
elif TARGET_BPOD_FIRMWARE_VERSION == "9":
    from pybpodapi.bpod.hardware.events.bpod0_7_5_fw9 import EventName
else:
    from pybpodapi.bpod.hardware.events.bpod0_7_9_fw20 import EventName
